#include "common_functions.h"

#include <iostream>
#include <iomanip>      // std::setw,
#include <sstream>
#include <fstream>
#include <string>
#include <vector>
#include <math.h>
#include <stdlib.h>
#include <time.h>

#include <Eigen/Dense>
#include <Eigen/Core>
#include <Eigen/SVD>

//#include <netcdfccp.h>

// GET TIME:
#if defined(_WIN32)
#include <Windows.h>

#elif defined(__unix__) || defined(__unix) || defined(unix) || (defined(__APPLE__) && defined(__MACH__))
#include <unistd.h> /* POSIX flags */
//#include <time.h>   /* clock_gettime(), time() */
#include <sys/time.h>   /* gethrtime(), gettimeofday() */

#if defined(__MACH__) && defined(__APPLE__)
#include <mach/mach.h>
#include <mach/mach_time.h>
#endif

#else
#error "Unable to define getRealTime( ) for an unknown OS."
#endif
//

#include "model1D.h"

using std::cout;
using std::endl;
using Eigen::Ref;


namespace common_functions
{
void test_function(model1D::Model1D some_model)
{

    cout << some_model.get_synthetic_field_filename() << endl;
    cout << some_model.get_weights_filename() << endl;
    cout << some_model.get_error_filename() << endl;
    cout << some_model.get_no_data() << endl;

}

void makeSyntheticModel(model1D::Model1D some_model, bool observed)
{
    std::stringstream mfil;
    
    if (observed)
        mfil << some_model.get_model_data_dir() << "/" << some_model.get_mfil_obs_filename();
    else
        mfil << some_model.get_mfil_temp_directory() << "/" << some_model.get_mfil_filename();

    //std::stringstream zlay;
    //std::stringstream mver;
    //zlay << some_model.get_water_depth() << "," << current_parameter_values(1) << "," << current_parameter_values(1) + current_parameter_values(2);
    //mver << some_model.get_water_resistivity() << "," << some_model.get_formation_resistivity() << "," << current_parameter_values(0) << "," << some_model.get_formation_resistivity();

    std::stringstream cmd_stream;
    
    cmd_stream << some_model.get_bin_dir() << "/elcardinal_R-1-8-1";
    cmd_stream << " par=" << some_model.get_par();
    cmd_stream << " freq=" << some_model.get_freq();
    //cmd_stream << " zlay=" << zlay.str();
    //cmd_stream << " mver=" << mver.str();
    //cmd_stream << " mhor=" << some_model.get_mhor();
    cmd_stream << " mfil=" << mfil.str();
    cmd_stream << " imod=1";
    cmd_stream << " zs=" << some_model.get_zs();
    cmd_stream << " zr=" << some_model.get_zr();
    cmd_stream << " r0=" << some_model.get_r0();
    cmd_stream << " dr=" << some_model.get_dr();
    cmd_stream << " rl=" << some_model.get_rl();
    cmd_stream << " sortyp=" << some_model.get_sortyp();
    cmd_stream << " camp=" << some_model.get_camp();
    cmd_stream << " alen=" << some_model.get_alen();
    cmd_stream << " azet=" << some_model.get_azet();
    cmd_stream << " tet=" << some_model.get_tet();
    cmd_stream << " method=" << some_model.get_method();
    cmd_stream << " acur=" << some_model.get_acur();
    cmd_stream << " comp=" << some_model.get_comp();
    cmd_stream << " gath=" << some_model.get_gath();
    cmd_stream << " tvi=" << some_model.get_tvi();
    cmd_stream << " psi=" << some_model.get_psi();
    cmd_stream << " x0=" << some_model.get_x0();
    cmd_stream << " y0=" << some_model.get_y0();

    if (observed)
        cmd_stream << " ofel=" << some_model.get_model_data_dir() << "/" << some_model.get_observed_field_filename();
    else
        cmd_stream << " ofel=" << some_model.get_temporary_dump_dir() << "/" << some_model.get_synthetic_field_filename();

    //std::cout << cmd_stream.str() << std::endl << std::endl;
    
    system(cmd_stream.str().c_str());

}

void makeWeights(model1D::Model1D some_model)
{
    double alpha = some_model.get_alpha();
    double eta_squared = some_model.get_eta_squared();
    std::string name = some_model.get_name();

    std::stringstream cmd_stream;
   

    if (name.substr(0,5) == "troll")
    {
        // Using invWeights3d
        cmd_stream << "~/SBLwiz_v1.26.5.19117_emgs/lib/invWeights3d_R-5-2-0 ";
        cmd_stream << "ifel_real=" << some_model.get_model_data_dir() << "/" << some_model.get_observed_field_filename() << " ";
        cmd_stream << "ofel=" << some_model.get_model_data_dir() << "/" << some_model.get_weights_filename() << " ";
        cmd_stream << "useNoise=false ";
        cmd_stream << "useMask=false ";
        cmd_stream << "minoffset=0 ";
        cmd_stream << "maxoffset=11000 ";
        cmd_stream << "scaling=1.0 ";
        cmd_stream << "freqExp=0.0 ";
        cmd_stream << "magRelativeErr=" << alpha << " ";
        cmd_stream << "snrLimit=1.0 ";
        cmd_stream << "noiseWt=1.0 ";
        cmd_stream << "magCutOff=1e-30 ";
        cmd_stream << "magErrOffset=no ";
        cmd_stream << "verbose=1";
    }
    else
    {
        // Using elmath
        cmd_stream << some_model.get_main_dir() << "/bin/elmath_R-0-16-1";
        cmd_stream << " ifel=" << some_model.get_model_data_dir() << "/" << some_model.get_observed_field_filename();
        cmd_stream << " op=mult fac=" << alpha << " | ";
        cmd_stream << some_model.get_main_dir() << "/bin/elmath_R-0-16-1 op=abs | ";
        cmd_stream << some_model.get_main_dir() << "/bin/elmath_R-0-16-1 op=power fac=2 | ";
        cmd_stream << some_model.get_main_dir() << "/bin/elmath_R-0-16-1 op=add fac=" << eta_squared << " | ";
        cmd_stream << some_model.get_main_dir() << "/bin/elmath_R-0-16-1 op=power fac=-1";
        cmd_stream << " ofel=" << some_model.get_model_data_dir() << "/" << some_model.get_weights_filename();
        system(cmd_stream.str().c_str());
    }
}

void makeMisfit(model1D::Model1D some_model)
{
    std::stringstream cmd_stream;

    cmd_stream << some_model.get_bin_dir() << "/elCalcWeiDiField_R-6-2-0"; 
    cmd_stream << " ifelObs=" << some_model.get_model_data_dir() << "/" << some_model.get_observed_field_filename();
    cmd_stream << " ifelSynth=" << some_model.get_temporary_dump_dir() << "/" << some_model.get_synthetic_field_filename();
    cmd_stream << " ifelWeights=" << some_model.get_model_data_dir() << "/" << some_model.get_weights_filename();
    cmd_stream << " ofel=" << some_model.get_temporary_dump_dir() << "/" << some_model.get_name() << "_tmp.nc";
    cmd_stream << " ErrorFile=" << some_model.get_temporary_dump_dir() << "/" << some_model.get_error_filename();
    
    //std::cout << cmd_stream.str() << std::endl;

    system(cmd_stream.str().c_str());


}

double misfitFunctionEvaluation(model1D::Model1D some_model, const Ref<Eigen::VectorXd>& current_parameter_values, int Np, int Nr, int Ng)
{
    resistivityFunctionEvaluation(some_model,current_parameter_values,"synthetic_temp", Np, Nr, Ng);
    makeSyntheticModel(some_model,false);
    makeMisfit(some_model);
    
    std::string misfit_filepath = some_model.get_temporary_dump_dir() + "/" + some_model.get_error_filename();

    std::fstream misfit_file(misfit_filepath.c_str(), std::ios::in);
    double misfit;
    misfit_file >> misfit;

    misfit_file.close();
    return misfit;
}

double resistivityFunctionEvaluation(model1D::Model1D some_model, const Ref<Eigen::VectorXd>& parameters, std::string file_location, int Np, int Nr, int Ng)
{
    // Creates a file which holds the values of the discretized resistivity
    // profile function. This is in turn loaded as input to the elcardinal
    // forward modeller.
    // file_location is one of the following: observed, synthetic_temp,
    // synthetic_keep
    int no_poly_terms = 1 + Np;
    int no_resistor_terms = Nr;
    int no_wavelet_terms = Ng;
    
    if (parameters.size() != (no_poly_terms + 2*no_resistor_terms + 3*no_wavelet_terms))
    {
        std::cout << "Error in parameter dimensions..." << std::endl;
        exit(1);
    }

    double resistor_thickness           = some_model.get_resistor_thickness(); 
    double z_water_min                  = some_model.get_z_water_min();
    double z_water_max                  = some_model.get_water_depth();
    double water_resistivity            = some_model.get_water_resistivity();
    
    double z_formation_min              = z_water_max;
    double z_formation_max              = some_model.get_z_formation_max();
    double initial_z_spacing            = some_model.get_initial_z_spacing();
    int z_formation_vector_length       = some_model.get_z_formation_vector_length();
    Eigen::VectorXd z_formation(z_formation_vector_length);
    dynamic_spacing(z_formation,z_formation_min,initial_z_spacing,z_formation_max,z_formation_vector_length);

    Eigen::VectorXd poly_parameters     = parameters.head(no_poly_terms);
    Eigen::VectorXd resistor_parameters = parameters.segment(no_poly_terms,2*no_resistor_terms);
    Eigen::VectorXd resistor_par_res    = resistor_parameters.head(no_resistor_terms);
    Eigen::VectorXd resistor_par_z      = resistor_parameters.tail(no_resistor_terms);
    Eigen::VectorXd wavelet_parameters  = parameters.tail(3*no_wavelet_terms);
    Eigen::VectorXd wavelet_par_beta    = wavelet_parameters.head(no_wavelet_terms);
    Eigen::VectorXd wavelet_par_mean      = wavelet_parameters.segment(no_wavelet_terms,no_wavelet_terms);
    Eigen::VectorXd wavelet_par_width     = wavelet_parameters.tail(no_wavelet_terms);
   

    // Polynomial term:
    Eigen::VectorXd polynomial_term = Eigen::MatrixXd::Zero(z_formation_vector_length,1);
    int i,j,k;
    double norm_const;
    for (i = 0; i < no_poly_terms; ++i)
    {
        norm_const = pow(z_formation_max - z_formation_min, (double) -i);
        for (j = 0; j < z_formation.size(); ++j)
        {
            polynomial_term(j) = polynomial_term(j) + poly_parameters(i)*norm_const*pow(z_formation(j) - z_formation_min,(double) i);
        }
    }

    // Thin resistor term:
    Eigen::VectorXd resistor_term = Eigen::MatrixXd::Zero(z_formation_vector_length,1);
    int nearest_upper = 0;
    int nearest_lower = 0;
    int temp_index = 0;
    for (i = 0; i < no_resistor_terms;++i)
    {
        nearest_upper = nearestIndex(z_formation,resistor_par_z(i));
        z_formation(nearest_upper) = resistor_par_z(i);
        
        nearest_lower = nearestIndex(z_formation,resistor_par_z(i) + some_model.get_resistor_thickness()); 
        z_formation(nearest_lower) = resistor_par_z(i) + some_model.get_resistor_thickness();
        
        // Double check this
        //for (i = 0; i < no_poly_terms; ++i)
        //{
        //    polynomial_term(nearest_upper + 1) += poly_parameters(i)*pow(z_formation(nearest_upper) - z_formation_min,(double) i);
        //    polynomial_term(nearest_lower) += poly_parameters(i)*pow(z_formation(nearest_lower) - z_formation_min,(double) i);
        //}

        temp_index = 0;
        while (nearest_upper + temp_index < nearest_lower)
        {
            resistor_term(nearest_upper + temp_index) += resistor_par_res(i);  // because of thickness of the layer
            temp_index++;
            //cout << nearest + temp_index << " < " << resistor_term.size() << endl;
            if (nearest_upper + temp_index >= resistor_term.size())
            {
                break;
            }
        }
    }

    // Smoothing wavelet term:
    Eigen::VectorXd wavelet_term = Eigen::MatrixXd::Zero(z_formation_vector_length,1);
    for (i = 0; i < no_wavelet_terms; ++i)
    {
        for (j = 0; j < z_formation_vector_length; ++j)
        {
            wavelet_term(j) += wavelet_par_beta(i)*pdf_gaussian(z_formation(j), wavelet_par_mean(i), wavelet_par_width(i));
        }
    }

    // formation resistivity profile
    Eigen::VectorXd formation_resistivity = Eigen::MatrixXd::Zero(z_formation_vector_length,1);
    formation_resistivity = polynomial_term + resistor_term + wavelet_term;

    Eigen::VectorXd resistivity_profile(z_formation_vector_length + 1);
    Eigen::VectorXd z_profile(z_formation_vector_length + 1);

    resistivity_profile(0) = water_resistivity;
    resistivity_profile.tail(z_formation_vector_length) = formation_resistivity;
    z_profile(0) = z_water_min;
    z_profile.tail(z_formation_vector_length) = z_formation;

    //for (i = 0; i < resistivity_profile.size();++i)
    //{
    //    if (resistivity_profile(i) < 1e-10)
    //        resistivity_profile(i) = 1e-10;
    //}

    bool write_mfil_ok = writeElcardinalMfil(some_model,z_profile,resistivity_profile,file_location);

    return resistivity_profile.minCoeff(); 
}

int nearestIndex(const Ref<Eigen::VectorXd>& vector, double value)
{
    // Returns the index in vector that holds the value closest to value
    
    double dist = 10000000.0;
    int temp_index = 0;
    double temp_dist = 0;

    for (int i = 0; i < vector.size(); ++i)
    {
        temp_dist = abs(vector(i) - value);
        if (temp_dist < dist)
        {
            dist = temp_dist;
            temp_index = i;
        }
    }
    return temp_index;
}

/*
void updateMinBetaValue(model1D::Model1D some_model, Ref<Eigen::VectorXd> min_parameter_values, const Ref<Eigen::VectorXd> parameter_values)
{
    // Is not in use... and does probably not work...
    int no_poly_terms = some_model.get_no_poly_terms();
    int no_resistor_terms = some_model.get_no_resistor_terms();
    int no_wavelet_terms = some_model.get_no_wavelet_terms();

    if (parameter_values.size() != (no_poly_terms + 2*no_resistor_terms + 3*no_wavelet_terms))
    {
        std::cout << "Error in parameter dimensions..." << std::endl;
        exit(1);
    }

    double resistor_thickness           = some_model.get_resistor_thickness(); 
    double z_formation_min              = atof( some_model.get_zr().c_str() );

    Eigen::VectorXd poly_parameters     = parameter_values.head(no_poly_terms);
    Eigen::VectorXd resistor_parameters = parameter_values.segment(no_poly_terms,2*no_resistor_terms);
    Eigen::VectorXd resistor_par_res    = resistor_parameters.head(no_resistor_terms);
    Eigen::VectorXd resistor_par_z      = resistor_parameters.tail(no_resistor_terms);
    Eigen::VectorXd wavelet_parameters  = parameter_values.tail(3*no_wavelet_terms);
    Eigen::VectorXd wavelet_par_beta    = wavelet_parameters.head(no_wavelet_terms);
    Eigen::VectorXd wavelet_par_mean    = wavelet_parameters.segment(no_wavelet_terms,no_wavelet_terms);
    Eigen::VectorXd wavelet_par_width   = wavelet_parameters.tail(no_wavelet_terms);
    
    double min_beta            = min_parameter_values[6]; 
    double poly_z_m = 0;
    double resistor_z_m = 0;
    double temp_val = 0;
    double sqrt_two_pi = 2.50662827463;
    double min_resistivity = 0.0;

    int i,j;

    for (i = 0; i < no_wavelet_terms; ++i)
    {
        for (j = 0; j < no_poly_terms; ++j)
        {
            poly_z_m += poly_parameters(j)*pow(wavelet_par_mean(i) - z_formation_min,(double) i);
        }

        for (j = 0; j < no_resistor_terms; ++j)
        {
            temp_val = ((resistor_par_z(j) < wavelet_par_mean(i)) && (resistor_par_z(j) + resistor_thickness > wavelet_par_mean(i))) ? resistor_par_res(j) : 0.0;
            resistor_z_m += temp_val;
        }
        min_beta(i) = sqrt_two_pi*wavelet_par_width(i)*(min_resistivity - poly_z_m - resistor_z_m);
        poly_z_m = 0;
        resistor_z_m = 0;
    }
    min_parameter_values.tail(3*no_wavelet_terms).head(no_wavelet_terms) = min_beta;
}
*/

void makeRandomSolution(model1D::Model1D some_model, Ref<Eigen::VectorXd> solution, const Ref<Eigen::VectorXd>& min_parameter_values, const Ref<Eigen::VectorXd> max_parameter_values, const Ref<Eigen::VectorXd>& initial_category_values, const Ref<Eigen::VectorXd>& category_standard_deviation, int initial_distribution, int Np, int Nr, int Ng)
{
    int i = 0;
   
    // constant term
    if (initial_distribution == 1)
    {
        solution(i) = min_parameter_values(0) + (max_parameter_values(0) - min_parameter_values(0))* (double) rand()/RAND_MAX;
    }
    else if (initial_distribution == 2)
    {
        solution(i) = initial_category_values(0) + generateGaussianNoise(category_standard_deviation(0)*category_standard_deviation(0));
        if (solution(i) < min_parameter_values(0))
            solution(i) = min_parameter_values(0);
        else if (solution(i) > max_parameter_values(0))
            solution(i) = max_parameter_values(0);
    }
    // higher order polynomial terms
    for (int j = 1;j < Np+1; ++j)
    {
        ++i;
        if (initial_distribution == 1)
        {
            solution(i) = min_parameter_values(1) + (max_parameter_values(1) - min_parameter_values(1))* (double) rand()/RAND_MAX;
        }
        else if (initial_distribution == 2)
        {
            solution(i) = initial_category_values(1) + generateGaussianNoise(category_standard_deviation(1)*category_standard_deviation(1));
            if (solution(i) < min_parameter_values(1))
                solution(i) = min_parameter_values(1);
            else if (solution(i) > max_parameter_values(1))
                solution(i) = max_parameter_values(1);
        }
    }

    // resistor terms
    //      resistivity
    for (int j = 1; j < Nr+1; ++j)
    {
        ++i;
        if (initial_distribution == 1)
        {
            solution(i) = min_parameter_values(2) + (max_parameter_values(2) - min_parameter_values(2))* (double) rand()/RAND_MAX;
        }
        else if (initial_distribution == 2)
        {
            solution(i) = initial_category_values(2) + generateGaussianNoise(category_standard_deviation(2)*category_standard_deviation(2));
            if (solution(i) < min_parameter_values(2))
                solution(i) = min_parameter_values(2);
            else if (solution(i) > max_parameter_values(2))
                solution(i) = max_parameter_values(2);
        }
    }
    //      depth
    for (int j = 1; j < Nr+1; ++j)
    {
        ++i;
        if (initial_distribution == 1)
        {
            solution(i) = min_parameter_values(3) + (max_parameter_values(3) - min_parameter_values(3))* (double) rand()/RAND_MAX;
        }
        else if (initial_distribution == 2)
        {
            solution(i) = initial_category_values(3) + generateGaussianNoise(category_standard_deviation(3)*category_standard_deviation(3));
            if (solution(i) < min_parameter_values(3))
                solution(i) = min_parameter_values(3);
            else if (solution(i) > max_parameter_values(3))
                solution(i) = max_parameter_values(3);
        }
    }
    
    // gaussian terms
    //      beta
    for (int j = 1; j < Ng+1; ++j)
    {
        ++i;
        if (initial_distribution == 1)
        {
            solution(i) = min_parameter_values(4) + (max_parameter_values(4) - min_parameter_values(4))* (double) rand()/RAND_MAX;
        }
        else if (initial_distribution == 2)
        {
            solution(i) = initial_category_values(4) + generateGaussianNoise(category_standard_deviation(4)*category_standard_deviation(4));
            if (solution(i) < min_parameter_values(4))
                solution(i) = min_parameter_values(4);
            else if (solution(i) > max_parameter_values(4))
                solution(i) = max_parameter_values(4);
        }
    }
    //      depth
    for (int j = 1; j < Ng+1; ++j)
    {
        ++i;
        if (initial_distribution == 1)
        {
            solution(i) = min_parameter_values(5) + (max_parameter_values(5) - min_parameter_values(5))* (double) rand()/RAND_MAX;
        }
        else if (initial_distribution == 2)
        {
            solution(i) = initial_category_values(5) + generateGaussianNoise(category_standard_deviation(5)*category_standard_deviation(5));
            if (solution(i) < min_parameter_values(5))
                solution(i) = min_parameter_values(5);
            else if (solution(i) > max_parameter_values(5))
                solution(i) = max_parameter_values(5);
        }
    }
    //      width
    for (int j = 1; j < Ng+1; ++j)
    {
        ++i;
        if (initial_distribution == 1)
        {
            solution(i) = min_parameter_values(6) + (max_parameter_values(6) - min_parameter_values(6))* (double) rand()/RAND_MAX;
        }
        else if (initial_distribution == 2)
        {
            solution(i) = initial_category_values(6) + generateGaussianNoise(category_standard_deviation(6)*category_standard_deviation(6));
            if (solution(i) < min_parameter_values(6))
                solution(i) = min_parameter_values(6);
            else if (solution(i) > max_parameter_values(6))
                solution(i) = max_parameter_values(6);
        }
    }
}


bool doubleArrayToFile(const double* array_in,size_t length, const std::string& file_path)
{
    // array_in should be declared as: double* array_in = new double[length];
    // fill it: for (int i = 0; i < length; i++) {array_in[i] = value(i)}
    // call function: doubleArrayToFile(array_in,length,"filename");
    std::ofstream os(file_path.c_str(), std::ios::binary | std::ios::out | std::ios::trunc);
    if (!os.is_open()) {return false;}
    os.write(reinterpret_cast<const char*>(array_in), std::streamsize(length*sizeof(double)));
    os.close();
    return true;
}

bool doubleArrayFromFile(double* array_out, size_t length, const std::string& file_path)
{
    // double* output_array = new double[length];
    // doubleArrayFromFile(output_array,length,"filename")
    //
    std::ifstream is(file_path.c_str(), std::ios::binary | std::ios::in);
    if (!is.is_open()) {return false;}
    is.read(reinterpret_cast<char*>(array_out), std::streamsize(length*sizeof(double)));
    is.close();
    return true;
}

bool vectorToFileDouble(const std::vector<double> &vec_in, const std::string& filepath)
{
    double* array_in = new double[vec_in.size()];
    for (int i = 0; i < vec_in.size(); ++i)
    {
        array_in[i] = vec_in[i];
    }
    bool to_file_ok = doubleArrayToFile(array_in, vec_in.size(), filepath);
    delete[] array_in;
    return to_file_ok;
}

bool vectorFromFileDouble(std::vector<double> &vec_out, const std::string& filepath)
{
    double* array_out = new double[vec_out.size()];
    bool from_file_ok = doubleArrayFromFile(array_out, vec_out.size(), filepath);
    for (int i = 0; i < vec_out.size(); ++i)
    {
        vec_out[i] = array_out[i];
    }
    delete[] array_out;
    return from_file_ok;
}

bool writeEigenVectorToBinaryFile(Ref<Eigen::VectorXd> vector,const std::string& filepath)
{
    std::ofstream f(filepath.c_str(), std::ios::out | std::ios::binary | std::ios::trunc);
    if (!f.is_open())
        return false;
    f.write(reinterpret_cast<char*>(vector.data()),std::streamsize(vector.size()*sizeof(double)));
    f.close();
    return true;
}

bool readEigenVectorFromBinaryFile(Ref<Eigen::VectorXd> vector,const std::string& filepath)
{
    std::ifstream f(filepath.c_str(), std::ios::in | std::ios::binary | std::ios::ate);
    if (!f.is_open())
        return false;
    std::streampos size;
    size = f.tellg();
    //vector.resize((int) size); // Remember to declare vector with the correct
    //number of elements.
    f.seekg(0, std::ios::beg);
    f.read(reinterpret_cast<char*>(vector.data()), size);
    f.close();
    return true;
}

bool writeEigenVectorToTextFile(const Ref<Eigen::VectorXd>& vector, const std::string& filepath)
{
    std::ofstream f(filepath.c_str(), std::ios::out | std::ios::trunc);
    if (!f.is_open())
        return false;
    f << vector;
    f.close();
    return true;
}

bool readEigenVectorFromTextFile(Ref<Eigen::VectorXd> vector, const std::string& filepath)
{
    std::ifstream f(filepath.c_str(), std::ios::in);    // Remember to declare vector with correct size
    if (!f.is_open())
        return false;
    int i = 0;
    int x;
    std::string line;
    double number;
    while (getline(f,line))
    {
        number = atof(line.c_str());
        vector(i) = number;
        i++;
    }
    f.close();
    return true;
}

bool writeElcardinalMfil(model1D::Model1D some_model,const Ref<Eigen::VectorXd>& z_values,const Ref<Eigen::VectorXd>& res_values,std::string file_location)
{
    int n_layer = res_values.size();
    int n_cmp = 1;
    double x0 = 0;
    double y0 = 0;
    double z0 = 0;
    double theta = 0;
    double drcmp = 0;

    std::string filepath;
    if (file_location == "synthetic_temp")
    {
        filepath = some_model.get_mfil_temp_directory() + "/" + some_model.get_mfil_filename();
    }
    else if (file_location == "synthetic_keep")
    {
        filepath = some_model.get_mfil_directory() + "/" + some_model.get_mfil_filename();
    }
    else if (file_location == "observed")
    {
        filepath = some_model.get_model_data_dir() + "/" + some_model.get_mfil_obs_filename();
    }
    else
    {
        filepath = file_location;
    }
    
    std::ofstream file(filepath.c_str(), std::ios::out | std::ios::trunc);
    if (!file.is_open())
        return false;
    file << std::setw(5) << n_layer << std::setw(5) << n_cmp << std::setw(5) << x0<< std::setw(5) << y0 << std::setw(5) << z0 << std::setw(5) << theta << std::setw(5) << drcmp << "\n";
    for (int i = 0; i < z_values.size(); ++i)
    {
        file << std::setw(16) << res_values(i) << std::setw(16) << res_values(i) << std::setw(16) << z_values(i) << "\n";
    }

    file.close();
}

void splitDouble(double d, int presision, int &integer_part, int &fractional_part)
{
    integer_part = (int)d;
    fractional_part = (int)( (d - integer_part)*std::pow(10,presision)+0.5);
}

void printDoubleVector(const std::vector<double> &vec)
{
    for (int i = 0; i < vec.size(); ++i)
        std::cout << vec[i] << endl;
}

void pushBackDoubleEigenMatrix(Ref<Eigen::MatrixXd> mat, const Ref<Eigen::VectorXd>& vec)
{
    int m = mat.rows();
    int n = mat.cols();

    for (int i = 0; i < m-1; ++i)
    {
        for (int j = 0; j < n; ++j)
        {
            mat(i,j) = mat(i+1,j);
        }
    }

    for (int j = 0; j < n; ++j)
        mat(m-1,j) = vec(j);
}

void pushBackDoubleEigenVector(Ref<Eigen::VectorXd> v, double val)
{
    // Appends val to the last entry in v, shifting every element one place up,
    // throws away first element

    int n = v.size();

    for (int i = 0; i < n-1; ++i)
        v(i) = v(i+1);
    v(n-1) = val;
}

// Math related functions

double sumSquaresDouble(const std::vector<double> &v)
{
    double result = 0.0;
    for (unsigned int i = 0; i < v.size(); ++i)
    {
        result += (v[i]*v[i]);
    }
    return result;
}

double pdf_gaussian(double x, double mu, double width)
{
    static const double inv_sqrt_2pi_var = 0.3989422804014327/width;
    double a = (x - mu) / width;

    return inv_sqrt_2pi_var*exp(-0.5*a*a);
}

double linear_interpolation(double f_a, double f_b, double a, double b, double x)
{
    return f_a + (f_b - f_a)*(x - a)/(b - a);
}

double logarithm_sum(const Ref<Eigen::VectorXd>& vector)
{
    return log(vector.sum());
}

void dynamic_spacing(Ref<Eigen::VectorXd> z_vector, double z_min, double dz_0, double z_max, int N)
{
    // adjust the z_vector, which is declared with length N, to a vector
    // containing increasingly spaced z-coordiantes sunch that z_vector[0] =
    // z_min, z_vector[1] = z_min + dz_0 and z_vector[N-1] = z_max.

    double increment_constant = 2.0/(((double)N-2.0)*((double)N-1.0))*(z_max - z_min - ((double)N-1.0)*dz_0);
    z_vector(0) = z_min;
    int n;

    for (n = 1; n < N ; ++n)
    {
        z_vector(n) = z_min + n*dz_0 + (n-1)*n/2*increment_constant;
    }


}

void multivariateGaussian(Ref<Eigen::VectorXd> x, const Ref<Eigen::VectorXd>& mu, const Ref<Eigen::MatrixXd>& covar)
{
    // Create a multivariate normal distributed vector x ~ gaussian_n (mu,covar)
    // x = mu + Az
    // where    mu is the mean vector
    //          z is standard gaussian distributed vector
    //          A = V*L^1/2, where covar = V*L*V^T

    int i,j;
    int n = x.size();
   
    // Create standard normal distributed vector z
    Eigen::VectorXd z(n);
    for (i = 0; i < n; ++i)
        z(i) = generateGaussianNoise(1.0);

    // Calculate A = V*L^1/2
    Eigen::SelfAdjointEigenSolver<Eigen::MatrixXd> es(covar);
    Eigen::VectorXd eigen_values = es.eigenvalues(); // L (as vector)
    Eigen::MatrixXd eigen_vectors = es.eigenvectors(); // V
    Eigen::MatrixXd sqrt_eigen_values = Eigen::MatrixXd::Zero(n,n);
    for (j = 0; j < n; ++j)
    {   
        sqrt_eigen_values(j,j) = sqrt(eigen_values(j));
    }

    Eigen::MatrixXd A = eigen_vectors*sqrt_eigen_values;
    
    x = mu + A*z;
}

double generateGaussianNoise(const double& variance)
{
    // Samples a N(0,variance) realization using the Box-Muller transform
    // From the wikipedia site on Box-Muller transform

    static bool hasSpare = false;
    static double rand1, rand2;
    double two_pi = 6.2831853071795864769252866;

    if (hasSpare)
    {
        hasSpare = false;
        return sqrt(variance*rand1)*sin(rand2);
    }

    hasSpare = true;

    rand1 = (double)rand()/((double)RAND_MAX);
    if (rand1 < 1e-100) rand1 = 1e-100;
    rand1 = -2*log(rand1);
    rand2 = (double)rand()/((double)RAND_MAX)*two_pi;

    return sqrt(variance*rand1)*cos(rand2);
}

double normalDistributionSample(double mean, double standard_deviation)
{
    return mean + generateGaussianNoise(standard_deviation*standard_deviation);
}

double sample_variance(const Ref<Eigen::VectorXd>& x)
{
    size_t n = x.size();
    double sample_mean_x = x.sum()/((double)n);
    double sample_variance_x = 0;
    for (size_t i = 0; i < n; ++i)
        sample_variance_x += (x(i) - sample_mean_x)*(x(i) - sample_mean_x);

    return sample_variance_x/((double)n - 1);
}


double correlation(const Ref<Eigen::VectorXd>& x, const Ref<Eigen::VectorXd>& y)
{
    size_t n = x.size();
    double Ex, Ey, Exy, Exx, Eyy;

    for (size_t i = 0; i < n; ++i)
    {
        Ex  += x(i);
        Ey  += y(i);
        Exy += x(i)*y(i);
        Exx += x(i)*x(i);
        Eyy += y(i)*y(i);
    }

    return ((double)n*Exy - Ex*Ey)/(sqrt((n*Exx - Ex*Ex)*(n*Eyy - Ey*Ey)));
}

double sgn_double(const double& val)
{
    return (0 < val) - (val < 0);
}

// REAL TIME CALCULATOR


/*
 * Author:  David Robert Nadeau
 * Site:    http://NadeauSoftware.com/
 * License: Creative Commons Attribution 3.0 Unported License
 *          http://creativecommons.org/licenses/by/3.0/deed.en_US
 */

/**
 * Returns the real time, in seconds, or -1.0 if an error occurred.
 *
 * Time is measured since an arbitrary and OS-dependent start time.
 * The returned real time is only useful for computing an elapsed time
 * between two calls to this function.
 */
double getRealTime()
{
#if defined(_WIN32)
    FILETIME tm;
    ULONGLONG t;
#if defined(NTDDI_WIN8) && NTDDI_VERSION >= NTDDI_WIN8
/* Windows 8, Windows Server 2012 and later. ---------------- */
    GetSystemTimePreciseAsFileTime( &tm );
#else
/* Windows 2000 and later. ---------------------------------- */
    GetSystemTimeAsFileTime( &tm );
#endif
    t = ((ULONGLONG)tm.dwHighDateTime << 32) | (ULONGLONG)tm.dwLowDateTime;
    return (double)t / 10000000.0;

#elif (defined(__hpux) || defined(hpux)) || ((defined(__sun__) || defined(__sun) || defined(sun)) && (defined(__SVR4) || defined(__svr4__)))
/* HP-UX, Solaris. ------------------------------------------ */
    return (double)gethrtime( ) / 1000000000.0;

#elif defined(__MACH__) && defined(__APPLE__)
/* OSX. ----------------------------------------------------- */
    static double timeConvert = 0.0;
    if ( timeConvert == 0.0 )
    {
       mach_timebase_info_data_t timeBase;
       (void)mach_timebase_info( &timeBase );
       timeConvert = (double)timeBase.numer /(double)timeBase.denom / 1000000000.0;
    } 
    return (double)mach_absolute_time( ) * timeConvert;

#elif defined(_POSIX_VERSION)
/* POSIX. --------------------------------------------------- */
#if defined(_POSIX_TIMERS) && (_POSIX_TIMERS > 0)
{
    struct timespec ts;
#if defined(CLOCK_MONOTONIC_PRECISE)
/* BSD. ----------------------------------------------------- */
    const clockid_t id = CLOCK_MONOTONIC_PRECISE;
#elif defined(CLOCK_MONOTONIC_RAW)
/* Linux. --------------------------------------------------- */
    const clockid_t id = CLOCK_MONOTONIC_RAW;
#elif defined(CLOCK_HIGHRES)
/* Solaris. ------------------------------------------------- */
    const clockid_t id = CLOCK_HIGHRES;
#elif defined(CLOCK_MONOTONIC)
/* AIX, BSD, Linux, POSIX, Solaris. ------------------------- */
    const clockid_t id = CLOCK_MONOTONIC;
#elif defined(CLOCK_REALTIME)
/* AIX, BSD, HP-UX, Linux, POSIX. --------------------------- */
    const clockid_t id = CLOCK_REALTIME;
#else
    const clockid_t id = (clockid_t)-1; /* Unknown. */
#endif /* CLOCK_* */
    if ( id != (clockid_t)-1 && clock_gettime( id, &ts ) != -1 )
        return (double)ts.tv_sec + (double)ts.tv_nsec / 1000000000.0;
    /* Fall thru. */
}
#endif /* _POSIX_TIMERS */

/* AIX, BSD, Cygwin, HP-UX, Linux, OSX, POSIX, Solaris. ----- */
    struct timeval tm;
    gettimeofday( &tm, NULL );
    return (double)tm.tv_sec + (double)tm.tv_usec / 1000000.0;
#else
    return -1.0;        /* Failed. */
#endif
}

int noDataPoints(model1D::Model1D some_model)
{
    // Reads the .nc-file containing the weights, and count number of non-zero
    // elements

    // Comment out until netcdf bug is fixed...
    //int no_data = some_model.get_no_data(); // number of data w/o masking
    //double weights[2*no_data][1];
    //
    //std::string filepath = some_model.get_model_data_dir() + "/" + some_model.get_weights_filename();
    //
    //NcFile dataFile(filepath,NcFile::ReadOnly);
    //
    //if (!dataFile.is_valid())
    //    return NC_ERR;
    //
    //NcVar *data = dataFile.get_var("emf");
    //
    //data -> get(&weights[0][0], no_data, 1);
    //
    //int no_data_points = 0;
    //
    //for (int i = 0; i < 2*no_data; ++i)
    //{
    //    if (weights[i][0] > 0.0)
    //        no_data_points++;
    //}
    
    std::string name = some_model.get_name();
    int no_data_points = 1;

    if (name.substr(0,9) == "troll_011")
        no_data_points = 144;
    else if (name.substr(0,9) == "troll_012")
        no_data_points = 190;
    else if (name.substr(0,9) == "troll_021")
        no_data_points = 190;
    else if (name.substr(0,9) == "troll_022")
        no_data_points = 190;
    else
        no_data_points = 273;
    
    
    return no_data_points;
}

double unifRand(double min, double max, bool log_uniform)
{
    double rand01 = (double) rand() / RAND_MAX;
    double y;
    double log_y;
    if (!log_uniform)
        y = min + (max - min)*rand01;
    else
    {
        log_y = log(min) + (log(max) - log(min))*rand01;
        y = exp(log_y);
    }
    return y;
}

int intUnifRand(int min, int max)
{
    return (min + (rand() % (int) (max - min + 1)));
}

void copyAndInsert(Ref<Eigen::VectorXd> new_vector, const Ref<Eigen::VectorXd>& old_vector, int index, double value)
{
    // equivalent to inserting value into index in old_vector

    int N_new = new_vector.size();
    int k = 0;
    for (int i = 0; i < N_new; ++i)
    {
        if (i == index)
            new_vector(i) = value;
        else
        {
            new_vector(i) = old_vector(k);
            k++;
        }
    }
}

void copyAndDelete(Ref<Eigen::VectorXd> new_vector, const Ref<Eigen::VectorXd>& old_vector, int index)
{
    int N_old = old_vector.size();
    int k = 0;
    for (int i = 0; i < N_old; ++i)
    {
        if (i != index)
        {
            new_vector(k) = old_vector(i);
            k++;
        }
    }

}

int categoryIndex(int index, int Np, int Nr, int Ng)
{
    int category_index;
    if (index == 0)
        category_index = 0;
    else if ((1 <= index) && (index <= Np) && (Np != 0))
        category_index = 1;
    else if ((Np < index) && (index <= Np+Nr) && (Nr != 0))
        category_index = 2;
    else if ((Np+Nr < index) && (index <= Np+2*Nr) && (Nr != 0))
        category_index = 3;
    else if ((Np+2*Nr < index) && (index <= Np+2*Nr+Ng) && (Ng != 0))
        category_index = 4;
    else if ((Np+2*Nr+Ng < index) && (index <= Np+2*Nr+2*Ng) && (Ng != 0))
        category_index = 5;
    else if ((Np+2*Nr+2*Ng < index) && (index <= Np+2*Nr+3*Ng) && (Ng != 0))
        category_index = 6;

    return category_index;
}

void inverseModelIndex(int model_index, int& np, int& nr, int& ng, int max_Np, int max_Nr, int max_Ng)
{
    ng = model_index % (max_Ng + 1);
    nr = ((model_index - ng) % ((max_Ng + 1)*(max_Nr + 1))) / (max_Ng + 1);
    np = ((model_index - (max_Ng + 1)*nr - ng) % ((max_Ng + 1)*(max_Nr + 1)*(max_Np + 1))) / ((max_Ng + 1)*(max_Nr + 1));
}

double minPolynomialValue(model1D::Model1D some_model, int order)
{
    double z_max = some_model.get_z_formation_max();
    double z_wd = some_model.get_water_depth();
    double k = some_model.get_min_poly_contribution();
    double c_min = k / pow(z_max - z_wd,(double) order);
    return c_min;
}

double maxPolynomialValue(model1D::Model1D some_model, int order)
{
    double z_max = some_model.get_z_formation_max();
    double z_wd = some_model.get_water_depth();
    double K = some_model.get_max_poly_contribution();
    double c_max = K / pow(z_max - z_wd,(double) order);
    return c_max;
}

double meanOfBestPartOfVector(const Ref<Eigen::VectorXd>& some_vector, double percent)
{
    // return the mean of the percent percent lowest values in some vector
    double fraction = percent / 100.0;
    int end = (int) std::max(1, static_cast<int>( round(some_vector.size()*fraction) ));
    int i = 0;
    double sum = 0;
    Eigen::VectorXd sorted_vector = some_vector;
    sortEigenVector(sorted_vector);
    for (int j = 0; j < end; ++j)
    {
        ++i;
        sum += sorted_vector(j);
    }
    return sum / (double) i;
}

void sortEigenVector(Ref<Eigen::VectorXd> some_vector)
{
    // selection sort
    int N = some_vector.size();
    int i,j;
    int min_index;
    double temp;
    for (j = 0; j < N-1; ++j)
    {
        min_index = j;
        for (i = j + 1; i < N; ++i)
        {
            if (some_vector(i) < some_vector(min_index))
                min_index = i;
        }
        if (min_index != j)
        {
            temp = some_vector(j);
            some_vector(j) = some_vector(min_index);
            some_vector(min_index) = temp;
        }
    }
}

void cumulativeFitness(const Ref<Eigen::VectorXd>& vector, Ref<Eigen::VectorXd> accumulated_vector)
{
    // modify the accumulated_vector, which is of size one larger than vector,
    // to contain accumulated values of fitness values of vector.

    Eigen::VectorXd norm_fitness = vector;
    int i;
    for (i = 0; i < vector.size(); ++i)
        norm_fitness(i) = (vector.maxCoeff() - vector(i))/(vector.maxCoeff() - vector.minCoeff());

    accumulate(norm_fitness,accumulated_vector);
    double max_ = accumulated_vector.maxCoeff();
    for (i = 0; i < accumulated_vector.size(); ++i)
        accumulated_vector(i) = accumulated_vector(i) / max_;

}

void accumulate(const Ref<Eigen::VectorXd>& vector, Ref<Eigen::VectorXd> accumulated_vector)
{
    accumulated_vector(0) = 0.0;
    double temp_sum = 0.0;
    int i,j;

    for (i = 0; i < vector.size(); ++i)
    {
        for (j = 0; j < i+1; ++j)
        {
            temp_sum += vector(j);
        }
        accumulated_vector(i+1) = temp_sum;
        temp_sum = 0.0;
    }

}


}


