#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <numeric>      // std::accumulate
#include <vector>       // std::vector
#include <algorithm>    // std::transform
#include <math.h>       // log, pow, exp, sqrt

#include <Eigen/Dense>  // compile as follows: prompt $ g++ -I ~/CPP/libraries filename.cpp -o filename
#include <Eigen/Core>
#include <Eigen/SVD>

#include "model1D.h"
#include "common_functions.h" 
#include "simulated_annealing.h"

using std::cout;
using std::endl;
using Eigen::Ref;



int main()
{
    model1D::Model1D some_model;
    simulated_annealing::SimulatedAnnealing some_run;

    // Choose values for the parameters most frequently changed.

    std::string name = "troll_022";                       
    
    // Modelling parameters for troll samples is set here, if not troll, default
    // constructor values is used
    bool noise = false;

    double zs;
    double zr;
    double x0;
    double y0;
    bool intow;
    double r0;
    double dr;
    double rl;
    double psi;
    double tet;
    double water_depth;
    double z_formation_max;
    double water_resistivity;

    if (name == "troll_011")                           // troll_001 | 011     | 012    || troll_002   | 021      | 022
    {               
        zs = 325.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 295.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 524634.6;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6734338.9;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = true;                                  //           | true    | false  ||             | true     | false
        r0 = -10000;                                   //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = -1000;                                    //           | -1000   | 10000  ||             | -1000    | 10000
        psi = 180.0;                                   // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 0.0;                                     // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 325.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
    }
    else if (name == "troll_012")
    {
        zs = 325.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 295.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 524634.6;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6734338.9;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = false;                                 //           | true    | false  ||             | true     | false
        r0 = 1000;                                     //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = 10000;                                    //           | -1000   | 10000  ||             | -1000    | 10000
        psi = 180.0;                                   // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 0.0;                                     // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 325.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
    }
    else if (name == "troll_021")
    {
        zs = 333.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 303.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 523237.3;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6743107.7;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = true;                                  //           | true    | false  ||             | true     | false
        r0 = -10000;                                   //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = -1000;                                    //           | -1000   | 10000  ||             | -1000    | 10000
        psi = 0.0;                                     // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 180.0;                                   // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 333.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
    }
    else if (name == "troll_022")
    {
        zs = 333.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 303.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 523237.3;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6743107.7;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = false;                                 //           | true    | false  ||             | true     | false
        r0 = 1000;                                     //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = 10000;                                    //           | -1000   | 10000  ||             | -1000    | 10000
        psi = 0.0;                                     // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 180.0;                                   // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 333.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
    }
    else if (name == "synthetic_001")
    {
        zs = 300;                                   
        zr = 329.99;                                    
        x0 = 0.0; 
        y0 = 0.0;                           
        intow = false;                           
        r0 = 1000;                                                        
        dr = 100;                              
        rl = 10000;                           
        psi = 90.0;                           
        tet = 0.0;                        
        water_depth = 330.0;               
        z_formation_max = 2500.0;         
        water_resistivity = 0.271;       
    }
    else if (name == "synthetic_002")
    {
        zs = 470;                                   
        zr = 499.99;                                    
        x0 = 0.0; 
        y0 = 0.0;                           
        intow = false;                           
        r0 = 1000;                                                        
        dr = 100;                              
        rl = 10000;                           
        psi = 90.0;                           
        tet = 0.0;                        
        water_depth = 500.0;               
        z_formation_max = 2500.0;         
        water_resistivity = 0.3125;       
    }
    else
    {
        zs = 333.59;                                   
        zr = 303.6;                                    
        x0 = 523237.3; 
        y0 = 6743107.7;                           
        intow = false;                           
        r0 = 1000;                                                        
        dr = 100;                              
        rl = 10000;                           
        psi = 0.0;                           
        tet = 180.0;                        
        water_depth = 333.6;               
        z_formation_max = 2500.0;         
        water_resistivity = 0.271;       
    }

    // Initial variable values (the free variables one is to optimize)
    int no_poly_terms = 2;
    int no_resistor_terms = 1;
    int no_wavelet_terms = 2;
    int no_parameters = no_poly_terms + 2*no_resistor_terms + 3*no_wavelet_terms;
    
    std::string result_type         = "final_results";  // "tests" or "final_results"
    std::string solution_method     = "dyndim";  // "fixed", "dyndim" or "transdim"

    double resistor_thickness       = 100.0;
    double min_misfit_value         = 150;
    bool max_iter_termination       = true;
    int max_iter                    = 20000;
    bool fn_value_termination       = true;
    bool reduce_step_length         = true;
    int epoch_const                 = 15;
    int num_iter_before_update      = epoch_const*no_parameters;
    double update_threshold         = 1e-3;
    int no_iter_history             = 10;
    int direction_type              = 1;
    double initial_temperature      = 10000;    //178790; 
    int temperature_schedule        = 3;        // 1: romeijn, 2: geometric, 3: lundy, 4: weighted average between 2 and 3
    double p_chi                    = 0.01;            
    double geometric_temp_const     = 0.9;
    double lundy_temp_const         = 5.0e-5;
    int poly_term_log_sample        = 0;        // 1: logarithmic uniform direction sampler, 2: logarithmic uniform step length sampler
    int new_solution_method         = 2;        // 1: old way, 2: new, 3: EMGS
    bool update_single_parameter    = true;
    bool update_temp_while_accept   = false;
    bool dyndim_lifo                = false;
    int min_Np                      = 1;        // minimum order of polynomial term, 0 means just the constant term
    int min_Nr                      = 1;        // minimum number of resistors
    int min_Ng                      = 1;        // minimum number of gaussfunctions
    int max_Np                      = 10;       // maximum order of polynomial term
    int max_Nr                      = 10;       // maximum number of resistors
    int max_Ng                      = 10;       // maximum number of resistors
    double p_nothing                = 0.7;      // probability of not changing number of parameters
    double p_in_Np                  = 0.05;      // probability of increase Np by one       
    double p_de_Np                  = 0.05;      // probability of decrease Np by one
    double p_in_Nr                  = 0.05;
    double p_de_Nr                  = 0.05;
    double p_in_Ng                  = 0.05;
    double p_de_Ng                  = 0.05;
    double min_poly_contribution    = 1e-8;     // also negative values..
    double max_poly_contribution    = 100;      // ''
    double best_model_prob          = 0.9;
    int initial_distribution        = 1;        // from which distribution to draw inital values. 1: uniform, 2: normal
    int seed                        = time(NULL); //1404829937  // seed to the random number generator

    srand(seed);                    // Remember to give run_id a unique number if the seed is reused
    rand();
    rand();
    rand();


    Eigen::VectorXd min_step_length_values(7);
    min_step_length_values <<   1e-3,            // 0. order polynomial term
                                1e-3,            // higher order polynomial terms
                                1e-2,            // resistor resistivity
                                1e-2,            // resistor depth
                                1e-2,            // gaussian beta
                                1e-2,            // gaussian depth
                                1e-2;            // gaussian width
    Eigen::VectorXd max_step_length_values(7);
    max_step_length_values <<   10.0,        // 0. order polynomial term
                                10.0,            // higher order polynomial terms
                                100.0,             // resistor resistivity
                                100.0,             // resistor depth
                                1000.0,            // gaussian beta
                                100.0,             // gaussian depth
                                100.0;             // gaussian width

    Eigen::VectorXd initial_category_values(7);
    initial_category_values << 1.0,
                               2.0,
                               10.0,
                               1500.0,
                               0.0,
                               1500.0,
                               100.0;

    Eigen::VectorXd category_standard_deviation(7);
    category_standard_deviation << 0.5,
                                   1.0,
                                   3.0,
                                   100.0,
                                   100.0,
                                   100.0,
                                   20.0;
   
    
    // Linear constraints:
    Eigen::VectorXd min_parameter_values(7);
    Eigen::VectorXd max_parameter_values(7);
    min_parameter_values    <<  0.0,                                    // constant term (0th order polynomial term)
                                -20.0,                                  // higher order polynomial terms
                                0.0,                                    // resistor resistivity 
                                water_depth,                            // resistor depth
                                -10000.0,                               // gaussian beta
                                water_depth,                            // gaussian depth
                                1.0;                                    // gaussian width
    max_parameter_values    <<  20.0,                                   // constant term (0th order polynomial term)
                                20.0,                                   // higher order polynomial terms
                                500.0,                                  // resistor resistivity
                                z_formation_max - resistor_thickness,   // reistor depth
                                10000.0,                                // gaussian beta
                                z_formation_max - resistor_thickness,   // gaussian depth
                                1000.0;                                 // gaussian width
    
    // Initial values (dimensions need to match parameter dimensions sat above)
    Eigen::VectorXd random_parameter_values = Eigen::MatrixXd::Zero(no_parameters,1);
    common_functions::makeRandomSolution(some_model,random_parameter_values,min_parameter_values,max_parameter_values,initial_category_values,category_standard_deviation,initial_distribution,no_poly_terms-1,no_resistor_terms,no_wavelet_terms);
   
    Eigen::VectorXd initial_parameter_values(no_parameters);
    initial_parameter_values << 0.3125,     
                                0.0,     
                                0.0,
                                random_parameter_values(3),     // random_parameter_values(3)
                                0.0,
                                0.0,
                                random_parameter_values(6),     // random_parameter_values(6)
                                random_parameter_values(7),     // random_parameter_values(7)
                                random_parameter_values(8),      // random_parameter_values(8)
                                random_parameter_values(9);      // random_parameter_values(9)
    
    initial_distribution = 1;                                    // for the rest of the run
   

    double log_sum = common_functions::logarithm_sum(initial_parameter_values);
    int run_id = rand();
    
    bool dynamic_dimensions;
    bool new_simulated_annealing;               // else, transdimensional simulated annealing, run_simulated_annealing is outdated

    if (solution_method == "fixed")
    {
        dynamic_dimensions          = false;
        new_simulated_annealing     = true;
    }
    else if (solution_method == "dyndim")
    {
        dynamic_dimensions          = true;
        new_simulated_annealing     = true;
    }
    else if (solution_method == "transdim")
    {
        dynamic_dimensions          = true;
        new_simulated_annealing     = false;
    }
    
    some_model.new_run(name, result_type, solution_method, noise,run_id,no_poly_terms,no_resistor_terms, no_wavelet_terms,log_sum);
    
    cout << "Solution file: " << endl << some_model.get_solution_filename() << endl;
    cout << "mfil file:     " << endl << some_model.get_mfil_filename() << endl;

    std::stringstream ss;
    std::string inputdir = some_model.get_initial_data_dir();
    std::string inputfile = some_model.get_initial_data_filename();
    ss << inputdir << "/" << inputfile;
    std::string initial_filepath = ss.str();

    //Eigen::VectorXd initial_data(4 + no_parameters);
    //initial_data(0) = noise;
    //initial_data(1) = no_poly_terms;
    //initial_data(2) = no_resistor_terms;
    //initial_data(3) = no_wavelet_terms;
    //initial_data.tail(no_parameters) = initial_parameter_values;
    //bool initial_data_write_ok = common_functions::writeEigenVectorToTextFile(initial_data,initial_filepath);
   
    std::ofstream initial_out_stream(initial_filepath.c_str(), std::ios::out | std::ios::trunc);
    if (!initial_out_stream.is_open())
        std::cout << "Unable to open file()" << endl;
    initial_out_stream << name << "\n";
    initial_out_stream << result_type << "\n";
    initial_out_stream << solution_method << "\n";
    initial_out_stream << noise << "\n";
    initial_out_stream << dynamic_dimensions << "\n";
    initial_out_stream << no_poly_terms << "\n";
    initial_out_stream << no_resistor_terms << "\n";
    initial_out_stream << no_wavelet_terms << "\n";
    initial_out_stream << initial_temperature << "\n";
    initial_out_stream << zs  << "\n";
    initial_out_stream << zr << "\n";
    initial_out_stream << x0 << "\n";
    initial_out_stream << y0 << "\n";
    initial_out_stream << intow << "\n";
    initial_out_stream << r0 << "\n";
    initial_out_stream << dr << "\n";
    initial_out_stream << rl << "\n";
    initial_out_stream << psi << "\n";
    initial_out_stream << tet << "\n";
    initial_out_stream << water_depth << "\n";
    initial_out_stream << z_formation_max << "\n";
    initial_out_stream << water_resistivity << "\n";
    initial_out_stream << resistor_thickness << "\n";
    
    int i;
    for (i = 0; i < no_parameters; ++i)
        initial_out_stream << initial_parameter_values[i] << "\n";
    
    initial_out_stream << min_misfit_value << "\n";
    initial_out_stream << max_iter_termination << "\n";
    initial_out_stream << max_iter  << "\n";
    initial_out_stream << fn_value_termination << "\n";
    initial_out_stream << reduce_step_length << "\n";
    initial_out_stream << epoch_const << "\n";
    initial_out_stream << num_iter_before_update << "\n";
    initial_out_stream << update_threshold << "\n";
    initial_out_stream << no_iter_history << "\n";
    initial_out_stream << direction_type << "\n";
    initial_out_stream << temperature_schedule << "\n";
    initial_out_stream << update_temp_while_accept << "\n";
    initial_out_stream << p_chi << "\n";
    initial_out_stream << geometric_temp_const << "\n";
    initial_out_stream << lundy_temp_const << "\n";
    initial_out_stream << poly_term_log_sample << "\n";
    initial_out_stream << new_solution_method << "\n";
    initial_out_stream << update_single_parameter << "\n";
    initial_out_stream << dyndim_lifo << "\n";
    initial_out_stream << min_Np << "\n";
    initial_out_stream << min_Nr << "\n";
    initial_out_stream << min_Ng << "\n";
    initial_out_stream << max_Np << "\n";
    initial_out_stream << max_Nr << "\n";
    initial_out_stream << max_Ng << "\n";
    initial_out_stream << p_nothing << "\n";
    initial_out_stream << p_in_Np << "\n";
    initial_out_stream << p_de_Np << "\n";
    initial_out_stream << p_in_Nr << "\n";
    initial_out_stream << p_de_Nr << "\n";
    initial_out_stream << p_in_Ng << "\n";
    initial_out_stream << p_de_Ng << "\n";
    initial_out_stream << min_poly_contribution << "\n";
    initial_out_stream << max_poly_contribution << "\n";
    
    for (i = 0; i < 7; ++i)
        initial_out_stream << min_step_length_values[i] << "\n";
    
    for (i = 0; i < 7; ++i)
        initial_out_stream << max_step_length_values[i] << "\n";
    
    initial_out_stream << best_model_prob << "\n";

    for (i = 0; i < 7; ++i)
        initial_out_stream << initial_category_values[i] << "\n";
    
    for (i = 0; i < 7; ++i)
        initial_out_stream << category_standard_deviation[i] << "\n";
    
    initial_out_stream << initial_distribution << "\n";
    initial_out_stream << seed << "\n";
    initial_out_stream.close();



    std::string min_parameter_values_filepath   = some_model.get_min_parameter_values_directory() + "/" + some_model.get_min_parameter_values_filename();
    std::string max_parameter_values_filepath   = some_model.get_max_parameter_values_directory() + "/" + some_model.get_max_parameter_values_filename();
    bool write_min_parameter_values = common_functions::writeEigenVectorToTextFile(min_parameter_values,min_parameter_values_filepath);
    bool write_max_parameter_values = common_functions::writeEigenVectorToTextFile(max_parameter_values,max_parameter_values_filepath);

    // Solve the optimization problem

    std::stringstream cmd_stream;
    if (new_simulated_annealing)
        cmd_stream << some_model.get_master_bin_dir() << "/dyndim_simulated_annealing " << initial_filepath << " " << no_parameters << " " << run_id;
    else
        cmd_stream << some_model.get_master_bin_dir() << "/transdim_simulated_annealing " << initial_filepath << " " << no_parameters << " " << run_id;
    system(cmd_stream.str().c_str());

    return 0;
}
