#include <iostream>
#include <string>
#include <sstream>
#include <cstdlib>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <numeric>      // std::accumulate
#include <vector>       // std::vector
#include <algorithm>    // std::transform
#include <math.h>       // log, pow, exp, sqrt

#include <Eigen/Dense>  // compile as follows: prompt $ g++ -I ~/CPP/libraries filename.cpp -o filename
#include <Eigen/Core>
#include <Eigen/SVD>

#include "model1D.h"
#include "common_functions.h" 


using std::cout;
using std::endl;
using Eigen::Ref;



int main()
{
    model1D::Model1D some_model;
    

    // Create model 
   
    std::string name = "synthetic_001";
    some_model.set_name(name);
    if (name == "troll_011" || name == "troll_021")
        some_model.set_intow(true);
    else if (name == "troll_012" || name == "troll_022")
        some_model.set_intow(false);
    double intow = some_model.get_intow();


    bool noise = false;
    int no_poly_terms = 3;
    int no_resistor_terms = 2;
    int no_wavelet_terms = 2;
    int no_parameters = no_poly_terms + 2*no_resistor_terms + 3*no_wavelet_terms;

    some_model.set_noise(noise);
    some_model.set_no_poly_terms(no_poly_terms);
    some_model.set_no_resistor_terms(no_resistor_terms);
    some_model.set_no_wavelet_terms(no_wavelet_terms);
    some_model.set_no_parameters(no_poly_terms,no_resistor_terms,no_wavelet_terms);

    Eigen::VectorXd observed_parameter_values(no_parameters);
    observed_parameter_values << 1.0,
                                0.0,
                                8.0,
                                20.0,
                                40.0,
                                1000.0,
                                1900.0,
                                -1000.0,
                                500.0,
                                1450.0,
                                500.0,
                                180.0,
                                200.0;
   
    double log_sum = common_functions::logarithm_sum(observed_parameter_values);
    srand(time(NULL));
    
    int run_id = rand();
    
    double d_r0;
    double d_dr;
    double d_rl;
    std::stringstream convert_r0(some_model.get_r0());
    std::stringstream convert_dr(some_model.get_dr());
    std::stringstream convert_rl(some_model.get_rl());
    if ( !(convert_r0 >> d_r0) )
        d_r0 = 0.0;
    if ( !(convert_dr >> d_dr) )
        d_dr = 0.0;
    if ( !(convert_rl >> d_rl) )
        d_rl = 0.0;
 
    some_model.set_directories(name,"tests","fixed");
    some_model.set_filenames(name, d_r0, d_dr, d_rl, noise, run_id, no_poly_terms, no_resistor_terms, no_wavelet_terms, log_sum);

    // write observed_parameter_values to file
    common_functions::writeEigenVectorToTextFile(observed_parameter_values,some_model.get_model_data_dir() + "/observed_parameter_values_" + name + ".txt");

    // make new model
    
    if (name.substr(0,5) == "troll")
    {
        // Truncate data
        std::stringstream cmd_stream_3;
        cmd_stream_3 << "~/SBLwiz_v1.26.5.19117_emgs/bin/elwin_R-3-5-2 ";
        cmd_stream_3 << "ifel=" << some_model.get_model_data_dir() << "/" << some_model.get_original_observed_field_filename() << " ";
        cmd_stream_3 << "ofel=" << some_model.get_model_data_dir() << "/" << some_model.get_observed_field_filename() << " ";
        cmd_stream_3 << "mode=offset ";
        if (intow)
        {
            cmd_stream_3 << "min=-10000 ";
            cmd_stream_3 << "max=-1000 ";
        }
        else
        {
            cmd_stream_3 << "min=1000 ";
            cmd_stream_3 << "max=10000 ";
        }
        cmd_stream_3 << "channels=ex1 ";
        cmd_stream_3 << "finc=0.25,0.75,1.25";
        system(cmd_stream_3.str().c_str());
        
        // Create weights
        // set values outside offset max and offset min to zero
        std::stringstream cmd_stream_1;
        cmd_stream_1 << "~/SBLwiz_v1.26.5.19117_emgs/lib/invWeights3d_R-5-2-0 ";
        cmd_stream_1 << "ifel_real=" << some_model.get_model_data_dir() << "/" << some_model.get_original_observed_field_filename() << " ";
        cmd_stream_1 << "ofel_real=none ";
        cmd_stream_1 << "ofel=" << some_model.get_model_data_dir() << "/" << some_model.get_temp_weights_filename() << " ";
        cmd_stream_1 << "useNoise=true ";
        cmd_stream_1 << "useMask=false ";
        if (name == "troll_011")
        {
            cmd_stream_1 << "minoffset=1500,1500,1000,0,0,0,0,0,0,0 ";
            cmd_stream_1 << "maxoffset=7000,6000,6000,11000,11000,11000,11000,11000,11000,11000 ";
        }
        else if (name == "troll_012")
        {
            cmd_stream_1 << "minoffset=1500,1500,1000,0,0,0,0,0,0,0 ";
            cmd_stream_1 << "maxoffset=9000,8000,6000,11000,11000,11000,11000,11000,11000,11000 ";
        }
        else if (name == "troll_021")
        {
            cmd_stream_1 << "minoffset=1500,1500,1000,0,0,0,0,0,0,0 ";
            cmd_stream_1 << "maxoffset=9000,8000,6000,11000,11000,11000,11000,11000,11000,11000 ";
        }
        else if (name == "troll_022")
        {
            cmd_stream_1 << "minoffset=1500,1500,1000,0,0,0,0,0,0,0 ";
            cmd_stream_1 << "maxoffset=9000,8000,6000,11000,11000,11000,11000,11000,11000,11000 ";
        }
        else
        {
            cmd_stream_1 << "minoffset=1500,1500,1000,0,0,0,0,0,0,0 ";
            cmd_stream_1 << "maxoffset=9000,8000,6000,11000,11000,11000,11000,11000,11000,11000 ";
        }
        cmd_stream_1 << "scaling=1.0 ";
        cmd_stream_1 << "freqExp=0.0 ";
        cmd_stream_1 << "magRelativeErr=0.02 "; 
        cmd_stream_1 << "snrLimit=20 ";
        cmd_stream_1 << "noiseWt=1.0 ";
        cmd_stream_1 << "magCutOff=1e-15 ";
        cmd_stream_1 << "magErrOffset=no ";
        cmd_stream_1 << "verbose=0";
        system(cmd_stream_1.str().c_str());
        
        // Truncate weights
        std::stringstream cmd_stream_2;
        cmd_stream_2 << "~/SBLwiz_v1.26.5.19117_emgs/bin/elwin_R-3-5-2 ";
        cmd_stream_2 << "ifel=" << some_model.get_model_data_dir() << "/" << some_model.get_temp_weights_filename() << " ";
        cmd_stream_2 << "ofel=" << some_model.get_model_data_dir() << "/" << some_model.get_weights_filename() << " ";
        cmd_stream_2 << "mode=offset ";
        if (intow)
        {
            cmd_stream_2 << "min=-10000 ";
            cmd_stream_2 << "max=-1000 ";
        }
        else
        {
            cmd_stream_2 << "min=1000 ";
            cmd_stream_2 << "max=10000 ";
        }
        cmd_stream_2 << "channels=ex1 ";
        cmd_stream_2 << "finc=0.25,0.75,1.25";
        system(cmd_stream_2.str().c_str());

    }
    else
    {
        double zs                           = 300.0;
        double zr                           = 329.99;
        double water_depth                  = zr + 0.01;
        
        some_model.set_zs(zs);
        some_model.set_zr(zr);
        some_model.set_water_depth(water_depth);

        common_functions::resistivityFunctionEvaluation(some_model,observed_parameter_values,"observed",no_poly_terms - 1,no_resistor_terms, no_wavelet_terms);
        common_functions::makeSyntheticModel(some_model,true);
        common_functions::makeWeights(some_model);
    }
    return 0;
}
