#ifndef MODEL1D_H
#define MODEL1D_H

#include <string>
#include <sstream>

#include <Eigen/Dense>
#include <Eigen/Core>

using Eigen::Ref;

namespace model1D
{

class Model1D
{

    public:
        // Constructors
        Model1D();
        Model1D(std::string name_, std::string result_type_, std::string solution_method_, double water_depth_, double water_resistivity_, std::string par_, double freq_[],int no_frequencies_, double mhor_, double zs_, double zr_, double r0_, double dr_, double rl_, std::string sortyp_, double camp_, double alen_, double azet_, double tet_, std::string method_, int acur_, std::string comp_, std::string gath_, std::string tvi_, double psi_, double x0_, double y0_, double zlay_obs_[], int zlay_obs_size_, double mver_obs_[], int mver_obs_size_, bool noise_, int run_id_, double min_misfit_value_, int no_poly_terms_, int no_resistor_terms_, int no_wavelet_terms_, double log_sum_);
        
        // Standard functions
        std::string DoubleArrayToString(double array[],int size);
        std::string DoubleToString(double value);
        std::string IntToString(int value);
        
        // Accessors and mutators
        void set_default();
        void set_model(std::string par_, double freq_[], int no_frequencies_, double mhor_, double zs_, double zr_, double r0_, double dr_, double rl_, std::string sortyp_, double camp_, double alen_, double azet_, double tet_, std::string method_, int acur_, std::string comp_, std::string gath_, std::string tvi_, double psi_, double x0_, double y0_, double zlay_obs_[], int zlay_obs_size_, double mver_obs_[], int mver_obs_size_);
        void set_filenames(std::string name_, double r0_, double dr_, double rl_, bool noise_, int run_id_, int no_poly_terms_, int no_resistor_terms_, int no_wavelet_terms_, double log_sum_);
        
        void set_intow(bool intow_);
        
        void set_zs(double zs_);
        void set_zr(double zr_);
        void set_x0(double x0_);
        void set_y0(double y0_);
        void set_tet(double tet_);
        void set_psi(double psi_);
        void set_r0(double r0_);
        void set_dr(double dr_);
        void set_rl(double rl_);

        //void set_model_variables(std::string name_);

        void set_directories(std::string name_, std::string result_type_, std::string solution_method_);
        void set_name(std::string name_);
        void set_water_depth(double water_depth_);
        void set_water_resistivity(double water_resistivity_);
        void set_no_data(double r0_, double dr_, double rl_, double no_frequencies_);
        void set_noise(bool noise_);
        void set_alpha(double alpha_);
        void set_eta_squared(double eta_squared_);
        void set_run_id(int run_id_);
        void set_min_misfit_value(double min_misfit_value_);
        void set_no_poly_terms(int no_poly_terms_);
        void set_no_resistor_terms(int no_resistor_terms_);
        void set_no_wavelet_terms(int no_wavelet_terms_);
        void set_no_parameters(int no_poly_terms_, int no_resistor_terms_, int no_wavelet_terms_);
        void set_log_sum(double log_sum_);
        void set_resistor_thickness(double resistor_thickness_);
        void set_z_water_min(double z_water_min_);
        void set_z_water_max(double z_water_max_);
        void set_z_formation_max(double z_formation_max_);
        void set_initial_z_spacing(double initial_z_spacing_);
        void set_z_formation_vector_length(double z_formation_vector_length_);

        void set_model_data_dir(std::string model_data_dir_);
        void set_mfil_obs_filename(std::string mfil_obs_filename_);
        void set_observed_field_filename(std::string observed_field_filename_);
        void set_weights_filename(std::string weights_filename_);
        void set_min_poly_contribution(double min_poly_contribution_);
        void set_max_poly_contribution(double max_poly_contribution_);



        void new_run(std::string name_, std::string result_type_, std::string solution_method_, bool noise_, int run_id_, int no_poly_terms_, int no_resistor_terms_, int no_wavelet_terms_, double log_sum_);

        std::string get_name() {return name;}

        bool get_intow() {return intow;}

        double get_min_misfit_value() {return min_misfit_value;}
        double get_water_depth() {return water_depth;}
        double get_water_resistivity() {return water_resistivity;}
        
        double get_resistor_thickness() {return resistor_thickness;}
        double get_z_water_min() {return z_water_min;} 
        double get_z_formation_max() {return z_formation_max;}
        double get_initial_z_spacing() {return initial_z_spacing;}
        int get_z_formation_vector_length() {return z_formation_vector_length;}
        double get_min_poly_contribution() {return min_poly_contribution;}
        double get_max_poly_contribution() {return max_poly_contribution;}

        std::string get_par() {return par;}
        std::string get_freq() {return freq;}
        std::string get_mhor() {return mhor;}
        std::string get_zs() {return zs;}
        std::string get_zr() {return zr;}
        std::string get_r0() {return r0;}
        std::string get_dr() {return dr;}
        std::string get_rl() {return rl;}
        std::string get_sortyp() {return sortyp;}
        std::string get_camp() {return camp;}
        std::string get_alen() {return alen;}
        std::string get_azet() {return azet;}
        std::string get_tet() {return tet;}
        std::string get_method() {return method;}
        std::string get_acur() {return acur;}
        std::string get_comp() {return comp;}
        std::string get_gath() {return gath;}
        std::string get_tvi() {return tvi;}
        std::string get_psi() {return psi;}
        std::string get_x0() {return x0;}
        std::string get_y0() {return y0;}
        std::string get_zlay_obs() {return zlay_obs;}
        std::string get_mver_obs() {return mver_obs;}

        bool get_noise() {return noise;}
        double get_alpha() {return alpha;}
        double get_eta_squared() {return eta_squared;}
        int get_run_id() {return run_id;}

        int get_no_receivers() {return no_receivers;}
        int get_no_sources() {return no_sources;}
        int get_no_frequencies() {return no_frequencies;}
        int get_no_field_components() {return no_field_components;}
        int get_no_data() {return no_data;}

        int get_no_poly_terms() {return no_poly_terms;}
        int get_no_resistor_terms() {return no_resistor_terms;}
        int get_no_wavelet_terms() {return no_wavelet_terms;}
        int get_no_parameters() {return no_parameters;}
        double get_log_sum() {return log_sum;}

        std::string get_observed_field_filename() {return observed_field_filename;}
        std::string get_original_observed_field_filename() {return original_observed_field_filename;}
        std::string get_synthetic_field_filename() {return synthetic_field_filename;}
        std::string get_weights_filename() {return weights_filename;}
        std::string get_temp_weights_filename() {return temp_weights_filename;}
        std::string get_error_filename() {return error_filename;}
        std::string get_initial_data_filename() {return initial_data_filename;}
        std::string get_mfil_filename() {return mfil_filename;}
        std::string get_mfil_obs_filename() {return mfil_obs_filename;}
        std::string get_parameter_filename() {return parameter_filename;}
        std::string get_min_parameter_values_filename() {return min_parameter_values_filename;}
        std::string get_max_parameter_values_filename() {return max_parameter_values_filename;}
        std::string get_solution_filename() {return solution_filename;}
        std::string get_variable_filename() {return variable_filename;}

        std::string get_main_dir() {return main_dir;}
        std::string get_bin_dir() {return bin_dir;}
        std::string get_master_bin_dir() {return master_bin_dir;}
        std::string get_temporary_dump_dir() {return temporary_dump_dir;}
        std::string get_model_data_dir() {return model_data_dir;}
        std::string get_method_data_dir() {return method_data_dir;}
        std::string get_initial_data_dir() {return initial_data_dir;}
        std::string get_mfil_temp_directory() {return mfil_temp_directory;}
        std::string get_mfil_directory() {return mfil_directory;}
        std::string get_parameter_directory() {return parameter_directory;}
        std::string get_min_parameter_values_directory() {return min_parameter_values_directory;}
        std::string get_max_parameter_values_directory() {return max_parameter_values_directory;}
        std::string get_siman_solution_dir() {return siman_solution_dir;}
        std::string get_siman_variable_dir() {return siman_variable_dir;}

    private:
        std::string name;

        bool intow;
        
        // Input to 1D elcardinal modeler
        std::string par;                    // "res": resistivity input, "con": conductivity input
        std::string freq;                   // comma seperated list of frequencies
        std::string mhor;                   // layer horizontal resistivity, mhor=0.0 is default and gives the same value as mver (vertical resistivity)
        std::string zs;                     // source depth
        std::string zr;                     // receiver depth
        std::string r0;                     // first offset
        std::string dr;                     // shot spacing (distance between offset)
        std::string rl;                     // last offset
        std::string sortyp;                 // source type. "el": electric, "ma": magnetic, "mm": magnetic moment
        std::string camp;                   // current amplitude
        std::string alen;                   // antenna horizontal length
        std::string azet;                   // antenna vertical length (azet < 0 : antenna directed down)
        std::string tet;                    // feather angle. angle between tow direction and antenna direction, clockwise
        std::string method;                 // integration tecnique, "quad" or "digf"
        std::string acur;                   // integration accuracy, lowest "1" to highest "9"
        std::string comp;                   // components to compute. "allhor","allver","exhy","exhyez","ex","exey","hy","hxhy","ez","hz"
        std::string gath;                   // 
        std::string tvi;                    // transverse isotropic modeling (anisotropic modeling: tvi="off")
        std::string psi;                    // towline direction, clockwise from north, (north=0, east=90.0 etc)
        std::string x0;                     // receiver easting
        std::string y0;                     // receiver northing
        std::string zlay_obs;               
        std::string mver_obs;
        //
        
        double min_misfit_value;
        double water_depth;
        double water_resistivity;
        bool noise;
        double alpha;                   // constant, relative error in the field amplitude measurement, instrument calibration inaccuracy
        double eta_squared;              // background noise
        
        int run_id;
        
        // Data gathering parameters
        int no_receivers;
        int no_sources;
        int no_frequencies;
        int no_field_components;
        int no_data;                   // Number of data (sources*receivers*frequencies*fields)
        
        // Resistivity profile parameters
        int no_poly_terms;
        int no_resistor_terms;
        int no_wavelet_terms;
        int no_parameters;
        double log_sum;                 // logarithm of the sum of the initial parameter values

        // Parameterization constants
        double resistor_thickness;    // Fixed target thickness (m)
        double z_water_min;           // 
        double z_formation_max;       // 
        double initial_z_spacing;     //
        int z_formation_vector_length;
        double min_poly_contribution;       // minimum value the polynomial terms can contribute with to the parameterization 
        double max_poly_contribution;       // maximum value the polynomial terms can contribute with to the parameterization 

        // File paths
        std::string observed_field_filename;
        std::string original_observed_field_filename;
        std::string synthetic_field_filename;
        std::string weights_filename;
        std::string temp_weights_filename;
        std::string error_filename;
        std::string initial_data_filename;
        std::string parameter_filename;
        std::string min_parameter_values_filename;
        std::string max_parameter_values_filename;
        std::string mfil_filename;
        std::string mfil_obs_filename;
        std::string solution_filename;
        std::string variable_filename;

        // Directories
        std::string main_dir;
        std::string bin_dir;
        std::string master_bin_dir;
        std::string temporary_dump_dir;
        std::string model_data_dir;
        std::string method_data_dir;
        std::string initial_data_dir;
        std::string parameter_directory;
        std::string min_parameter_values_directory;
        std::string max_parameter_values_directory;
        std::string mfil_temp_directory;
        std::string mfil_directory;
        std::string siman_solution_dir;
        std::string siman_variable_dir;
};
}

#endif
