#include <iostream>     // std::cout,std::cin,std::endl;
#include <string>       // std::string
#include <sstream>      // std::stringstream
#include <vector>       // std::vector<>
#include <time.h>       // time_t, struct tm, difftime, time, mktime 
#include <ctime>        // 
#include <cstdlib>      // RAND_MAX
#include <stdlib.h>     // srand, rand
#include <math.h>       // sqrt, exp, pow, log
#include <numeric>      // std::accumulate,
#include <algorithm>    // std::transform, std::max
#include <iomanip>      // std::setw,

#include <Eigen/Dense>
#include <Eigen/Core>
#include <Eigen/SVD>

#include <boost/lambda/lambda.hpp>
#include <boost/math/distributions/chi_squared.hpp>

#include "model1D.h"
#include "common_functions.h"
#include "simulated_annealing.h"

using boost::math::chi_squared;
using boost::math::quantile;
using std::cout;
using std::endl;
using Eigen::Ref;


int main()
{
    model1D::Model1D some_model;

    // Choose values for the parameters most frequently changed.

    std::string name = "troll_011";                       
    
    // Modelling parameters for troll samples is set here, if not troll, default
    // constructor values is used
    bool noise = false;

    double zs;
    double zr;
    double x0;
    double y0;
    bool intow;
    double r0;
    double dr;
    double rl;
    double psi;
    double tet;
    double water_depth;
    double z_formation_max;
    double water_resistivity;
    int no_data;

    if (name == "troll_011")                           // troll_001 | 011     | 012    || troll_002   | 021      | 022
    {               
        zs = 325.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 295.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 524634.6;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6734338.9;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = true;                                  //           | true    | false  ||             | true     | false
        r0 = -10000;                                   //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = -1000;                                    //           | -1000   | 10000  ||             | -1000    | 10000
        psi = 180.0;                                   // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 0.0;                                     // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 325.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
        no_data = 144;
    }
    else if (name == "troll_012")
    {
        zs = 325.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 295.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 524634.6;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6734338.9;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = false;                                 //           | true    | false  ||             | true     | false
        r0 = 1000;                                     //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = 10000;                                    //           | -1000   | 10000  ||             | -1000    | 10000
        psi = 180.0;                                   // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 0.0;                                     // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 325.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
        no_data = 190;
    }
    else if (name == "troll_021")
    {
        zs = 333.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 303.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 523237.3;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6743107.7;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = true;                                  //           | true    | false  ||             | true     | false
        r0 = -10000;                                   //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = -1000;                                    //           | -1000   | 10000  ||             | -1000    | 10000
        psi = 0.0;                                     // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 180.0;                                   // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 333.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
        no_data = 190;
    }
    else if (name == "troll_022")
    {
        zs = 333.59;                                   // 325.59    |  "      |  "     || 333.59      |          |
        zr = 303.6;                                    // 295.6     |  "      |  "     || 303.6       |          |
        x0 = 523237.3;                                 // 524634.6  |  "      |  "     || 523237.3    |          |
        y0 = 6743107.7;                                // 6734338.9 |  "      |        || 6743107.7   |          |
        intow = false;                                 //           | true    | false  ||             | true     | false
        r0 = 1000;                                     //           | -10000  | 1000   ||             | -10000   | 1000                          
        dr = 100;                                      //           | 100     | 100    ||             | 100      | 100
        rl = 10000;                                    //           | -1000   | 10000  ||             | -1000    | 10000
        psi = 0.0;                                     // 180.0 (south)  | "  |  "     || 0.0 (north) |          |
        tet = 180.0;                                   // 0.0       |  "      |  "     || 180.0       |          |
        water_depth = 333.6;                           // 325.6     |  "      |  "     || 333.6       |          |
        z_formation_max = 2500.0;                      // 2500.0    |  "      |  "     || 2500.0      |          |
        water_resistivity = 0.271;                     // 0.271     |  "      |  "     || 0.271       |          |
        no_data = 190;
    }
    else
    {
        zs = 333.59;                                   
        zr = 303.6;                                    
        x0 = 523237.3; 
        y0 = 6743107.7;                           
        intow = false;                           
        r0 = 1000;                                                        
        dr = 100;                              
        rl = 10000;                           
        psi = 0.0;                           
        tet = 180.0;                        
        water_depth = 333.6;               
        z_formation_max = 2500.0;         
        water_resistivity = 0.271;       
        no_data = 273;
    }
    int no_poly_terms = 4;
    int no_resistor_terms = 0;
    int no_wavelet_terms = 3;
    int no_parameters = no_poly_terms + 2*no_resistor_terms + 3*no_wavelet_terms;
   
    Eigen::VectorXd parameter_values(no_parameters);
    parameter_values << 0.643496,
                        2.1061,
                        0.846811,
                        -2.05732,
                        5184.79,
                        621.233,
                        -857.637,
                        1693.91,
                        939.776,
                        963.96,
                        57.272,
                        999.995,
                        101.758;
    
    int log_sum = 1;
    int run_id = 2;

    // Elcardinal input

    
    some_model.set_zs(zs);
    some_model.set_zr(zr);
    some_model.set_x0(x0);
    some_model.set_y0(y0);
    some_model.set_intow(intow);
    some_model.set_r0(r0);                          
    some_model.set_dr(dr);
    some_model.set_rl(rl);
    some_model.set_psi(psi);
    some_model.set_tet(tet);
    some_model.set_water_depth(water_depth);
    some_model.set_z_formation_max(z_formation_max);
    some_model.set_water_resistivity(water_resistivity);
    some_model.new_run(name,"tests","transdim",noise,run_id,no_poly_terms,no_resistor_terms, no_wavelet_terms,log_sum);

    double misfit = common_functions::misfitFunctionEvaluation(some_model, parameter_values, no_poly_terms - 1, no_resistor_terms, no_wavelet_terms);
    double RMS = sqrt(misfit/(double)no_data);
    
    common_functions::resistivityFunctionEvaluation(some_model,parameter_values,"synthetic_keep", no_poly_terms - 1, no_resistor_terms, no_wavelet_terms);

    cout << endl << "RMS = " << RMS << endl;
    
    //std::stringstream cmd_stream;
    //cmd_stream << "gnuplot -e ";
    //cmd_stream << "\"filename='" << some_model.get_mfil_filename() << "'\" ";
    //cmd_stream << some_model.get_mfil_directory() << "/" << some_model.get_mfil_filename();
    //system(cmd_stream.str().c_str());

    return 0;
}
