// Program used to test various implemented code

#include <iostream>
#include <string>
#include <sstream>
#include <vector>
#include <numeric>
#include <algorithm>
#include <math.h>
#include <cstdlib>
#include <stdlib.h>
#include <time.h>

#include "model1D.h"
#include "common_functions.h"
#include "simulated_annealing.h"


using std::cout;
using std::endl;

int main()
{
    int dimensions = 3;
    double array_initial[3] = {90.0,1900.0,120.0};
    std::vector<double> vector_initial(array_initial, array_initial + sizeof(array_initial)/sizeof(array_initial[0]));
    
    srand(time(NULL));
    int run_id = rand();

    double initial_sum = std::accumulate(vector_initial.begin(), vector_initial.end(), 0.0);
    double log_sum = log(initial_sum);
    
    cout << initial_sum << endl;
    cout << log_sum << endl;

    model1D::Model1D some_model;

    some_model.new_run(run_id,dimensions,log_sum);

    std::stringstream ss3;
    std::string inputdir = some_model.get_initial_data_dir();
    std::string inputfile = some_model.get_initial_data_filename();
    ss3 << inputdir << "/" << inputfile;
    std::string initial_filepath = ss3.str();

    bool initial_values_write_ok = common_functions::vectorToFileDouble(vector_initial, initial_filepath);
    
    cout << initial_filepath << endl;
    cout << initial_values_write_ok << endl;

    model1D::Model1D run_model;
    simulated_annealing::SimulatedAnnealing this_run;
   
    run_model.new_run(run_id,dimensions,log_sum);   // Prepare for new run

    std::stringstream initial_data_stream;
    std::string initial_data_dir = run_model.get_initial_data_dir();
    std::string initial_data_filename = run_model.get_initial_data_filename();
    initial_data_stream << initial_data_dir << "/" << initial_data_filename;
    std::string initial_filepathh = initial_data_stream.str();
    
    std::vector<double> initial_values(dimensions,0.0);
    bool initial_values_read_ok = common_functions::vectorFromFileDouble(initial_values, initial_filepathh);

    cout << initial_filepathh << endl;
    cout << initial_values_read_ok << endl;

    common_functions::printDoubleVector(initial_values);

    return 0;
}
