// Like dyndim_simulated_annealing.cpp but another approach

#include <iostream>     // std::cout,std::cin,std::endl;
#include <string>       // std::string
#include <sstream>      // std::stringstream
#include <vector>       // std::vector<>
#include <time.h>       // time_t, struct tm, difftime, time, mktime 
#include <ctime>        // 
#include <cstdlib>      // RAND_MAX
#include <stdlib.h>     // srand, rand
#include <math.h>       // sqrt, exp, pow, log
#include <numeric>      // std::accumulate,
#include <algorithm>    // std::transform, std::max
#include <iomanip>      // std::setw,

#include <Eigen/Dense>
#include <Eigen/Core>
#include <Eigen/SVD>

#include <boost/lambda/lambda.hpp>
#include <boost/math/distributions/chi_squared.hpp>

#include "model1D.h"
#include "common_functions.h"
#include "simulated_annealing.h"

using boost::math::chi_squared;
using boost::math::quantile;
using std::cout;
using std::endl;
using Eigen::Ref;

int main(int argc, char *argv[])
{
    double start_time, end_time;
    start_time = common_functions::getRealTime();

    // Command line arguments:
    std::string initial_filepath = argv[1];
    int no_parameters = atoi (argv[2]);
    int run_id = atoi (argv[3]);
    std::stringstream ss1;
    ss1 << run_id;
    std::string string_run_id = ss1.str();

    // Arguments from file
    std::string name;
    std::string result_type;
    std::string solution_method;
    bool noise;
    bool dynamic_dimensions;
    int no_poly_terms;
    int no_resistor_terms;
    int no_wavelet_terms;
    double initial_temperature;
    double zs;
    double zr;
    double x0;
    double y0;
    bool intow;
    double r0;                          
    double dr;
    double rl;
    double psi;
    double tet;
    double water_depth;
    double z_formation_max;
    double water_resistivity;
    double resistor_thickness;
    Eigen::VectorXd parameters(no_parameters);
    double min_misfit_value;
    bool max_iter_termination;
    int max_iter;
    bool fn_value_termination;
    bool reduce_step_length;
    int epoch_const;
    int num_iter_before_update;
    double update_threshold;
    int no_iter_history;
    int direction_type;
    int temperature_schedule;        // 1: romeijn, 2: geometric, 3: lundy
    bool update_temp_while_accept;
    double p_chi;            
    double geometric_temp_const;
    double lundy_temp_const;
    int poly_term_log_sample;        // 1: logarithmic uniform direction sampler, 2: logarithmic uniform step length sampler
    int new_solution_method;        // 1: old way, 2: new, 3: EMGS
    bool update_single_parameter;
    bool dyndim_lifo;
    int min_Np;        // minimum order of polynomial term, 0 means just the constant term
    int min_Nr;        // minimum number of resistors
    int min_Ng;        // minimum number of gaussfunctions
    int max_Np;       // maximum order of polynomial term
    int max_Nr;       // maximum number of resistors
    int max_Ng;       // maximum number of resistors
    double p_nothing;      // probability of not changing number of parameters
    double p_in_Np;      // probability of increase Np by one       
    double p_de_Np;      // probability of decrease Np by one
    double p_in_Nr;
    double p_de_Nr;
    double p_in_Ng;
    double p_de_Ng;
    double min_poly_contribution;
    double max_poly_contribution;
    Eigen::VectorXd min_step_length_values(7);
    Eigen::VectorXd max_step_length_values(7);
    double best_model_prob;
    Eigen::VectorXd initial_category_values(7);
    Eigen::VectorXd category_standard_deviation(7);
    int initial_distribution;
    int seed;

    std::ifstream initial_in_stream(initial_filepath.c_str(), std::ios::in);
    if (!initial_in_stream.is_open())
        std::cout << "Unable to open file()" << endl;
    initial_in_stream >> name;
    initial_in_stream >> result_type;
    initial_in_stream >> solution_method;
    initial_in_stream >> noise;
    initial_in_stream >> dynamic_dimensions;
    initial_in_stream >> no_poly_terms;
    initial_in_stream >> no_resistor_terms;
    initial_in_stream >> no_wavelet_terms;
    initial_in_stream >> initial_temperature;
    initial_in_stream >> zs;
    initial_in_stream >> zr;
    initial_in_stream >> x0;
    initial_in_stream >> y0;
    initial_in_stream >> intow;
    initial_in_stream >> r0;
    initial_in_stream >> dr;
    initial_in_stream >> rl;
    initial_in_stream >> psi;
    initial_in_stream >> tet;
    initial_in_stream >> water_depth;
    initial_in_stream >> z_formation_max;
    initial_in_stream >> water_resistivity;
    initial_in_stream >> resistor_thickness;
    int i;
    for (i = 0; i < no_parameters; ++i)
        initial_in_stream >> parameters[i];
    initial_in_stream >> min_misfit_value;
    initial_in_stream >> max_iter_termination;
    initial_in_stream >> max_iter;
    initial_in_stream >> fn_value_termination;
    initial_in_stream >> reduce_step_length;
    initial_in_stream >> epoch_const;
    initial_in_stream >> num_iter_before_update;
    initial_in_stream >> update_threshold;
    initial_in_stream >> no_iter_history;
    initial_in_stream >> direction_type;
    initial_in_stream >> temperature_schedule;
    initial_in_stream >> update_temp_while_accept;
    initial_in_stream >> p_chi;
    initial_in_stream >> geometric_temp_const;
    initial_in_stream >> lundy_temp_const;
    initial_in_stream >> poly_term_log_sample;
    initial_in_stream >> new_solution_method;
    initial_in_stream >> update_single_parameter;
    initial_in_stream >> dyndim_lifo;
    initial_in_stream >> min_Np;
    initial_in_stream >> min_Nr;
    initial_in_stream >> min_Ng;
    initial_in_stream >> max_Np;
    initial_in_stream >> max_Nr;
    initial_in_stream >> max_Ng;
    initial_in_stream >> p_nothing;
    initial_in_stream >> p_in_Np;
    initial_in_stream >> p_de_Np;
    initial_in_stream >> p_in_Nr;
    initial_in_stream >> p_de_Nr;
    initial_in_stream >> p_in_Ng;
    initial_in_stream >> p_de_Ng;
    initial_in_stream >> min_poly_contribution;
    initial_in_stream >> max_poly_contribution;
    
    for (i = 0; i < 7; ++i)
        initial_in_stream >> min_step_length_values[i];
    
    for (i = 0; i < 7; ++i)
        initial_in_stream >> max_step_length_values[i];

    initial_in_stream >> best_model_prob;

    for (i = 0; i < 7; ++i)
        initial_in_stream >> initial_category_values[i];
    
    for (i = 0; i < 7; ++i)
        initial_in_stream >> category_standard_deviation[i];

    initial_in_stream >> initial_distribution;
    initial_in_stream >> seed;
    
    initial_in_stream.close();
    
    // Model object
    model1D::Model1D run_model; 

    run_model.set_zs(zs);
    run_model.set_zr(zr);
    run_model.set_x0(x0);
    run_model.set_y0(y0);
    run_model.set_intow(intow);
    run_model.set_r0(r0);                          
    run_model.set_dr(dr);
    run_model.set_rl(rl);
    run_model.set_psi(psi);
    run_model.set_tet(tet);
    run_model.set_water_depth(water_depth);
    run_model.set_resistor_thickness(resistor_thickness);
    run_model.set_z_formation_max(z_formation_max);
    run_model.set_water_resistivity(water_resistivity);
    run_model.set_min_poly_contribution(min_poly_contribution);
    run_model.set_max_poly_contribution(max_poly_contribution);
    run_model.set_min_misfit_value(min_misfit_value);
    if ((parameters.size() != no_parameters) && (parameters.size() != (no_poly_terms + 2*no_resistor_terms + 3*no_wavelet_terms)))
        cout << "Something is wrong with parameter no_parameters." << endl;

    double log_sum = common_functions::logarithm_sum(parameters);
    run_model.new_run(name, result_type, solution_method, noise, run_id, no_poly_terms, no_resistor_terms, no_wavelet_terms, log_sum);

    int no_data_points = common_functions::noDataPoints(run_model);
    
    // Simulated annealing object    
    simulated_annealing::SimulatedAnnealing this_run;
    this_run.set_dynamic_dimensions(dynamic_dimensions);
    this_run.set_num_iter_before_update(num_iter_before_update);
    this_run.set_max_iter_termination(max_iter_termination);
    this_run.set_max_iter(max_iter);
    this_run.set_fn_value_termination(fn_value_termination);
    this_run.set_reduce_step_length(reduce_step_length);
    this_run.set_epoch_const(epoch_const);
    this_run.set_num_iter_before_update(num_iter_before_update);
    this_run.set_update_threshold(update_threshold);
    this_run.set_no_iter_history(no_iter_history);
    this_run.set_direction_type(direction_type);
    this_run.set_temperature_schedule(temperature_schedule);
    this_run.set_update_temp_while_accept(update_temp_while_accept);
    this_run.set_p_chi (p_chi);
    this_run.set_geometric_temp_const(geometric_temp_const);
    this_run.set_lundy_temp_const(lundy_temp_const);
    this_run.set_poly_term_log_sample(poly_term_log_sample);
    this_run.set_new_solution_method(new_solution_method);
    this_run.set_update_single_parameter(update_single_parameter);
    this_run.set_min_Np(min_Np);
    this_run.set_min_Nr(min_Nr);
    this_run.set_min_Ng(min_Ng);
    this_run.set_max_Np(max_Np);
    this_run.set_max_Nr(max_Nr);
    this_run.set_max_Ng(max_Ng);
    this_run.set_p_nothing(p_nothing);
    this_run.set_p_in_Np(p_in_Np);
    this_run.set_p_de_Np(p_de_Np);
    this_run.set_p_in_Nr(p_in_Nr);
    this_run.set_p_de_Nr(p_de_Nr);
    this_run.set_p_in_Ng(p_in_Ng);
    this_run.set_p_de_Ng(p_de_Ng);
    this_run.set_best_model_prob(best_model_prob);
    
    int Np = no_poly_terms - 1;
    int Nr = no_resistor_terms;
    int Ng = no_wavelet_terms;

    std::vector<double> all_misfit_values;           // std::vector because of push_back
    std::vector<double> latest_misfit_values;
    double sum_latest_misfit_values;
    Eigen::VectorXd direction = Eigen::MatrixXd::Zero(parameters.size(),1);

    // Linear constraints:
    std::string min_parameter_values_filepath   = run_model.get_min_parameter_values_directory() + "/" + run_model.get_min_parameter_values_filename();
    std::string max_parameter_values_filepath   = run_model.get_max_parameter_values_directory() + "/" + run_model.get_max_parameter_values_filename();
    Eigen::VectorXd min_parameter_values(no_parameters);
    Eigen::VectorXd max_parameter_values(no_parameters);
    common_functions::readEigenVectorFromTextFile(min_parameter_values,min_parameter_values_filepath);
    common_functions::readEigenVectorFromTextFile(max_parameter_values,max_parameter_values_filepath);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    double step_length = 0.0;
    int num_halves = 0;         // Number of times the max step length sample threshold is halved
   
    //cout << current_solution << endl;

    double min_resistivity_value = run_model.get_water_resistivity();
    double current_resistivity_min = 0.0;
    bool current_resistivity_is_smaller = true;
    
    // Create and initialize archive
    // archive is a vector of vectors (models), where each vector represent a combination
    // of (n_p, n_r, n_g) and is called a model. model 0 is (0,0,0) and the
    // final model is (max_Np, max_Nr, max_Ng).
    std::vector< std::vector<double> > archive;
    std::vector<double> temp_bin;
    Eigen::VectorXd temp_archive;
    std::vector< std::vector<double> > best_solutions;
    std::vector<double> best_misfit_values;
    Eigen::VectorXd current_solution = Eigen::MatrixXd::Zero(parameters.size(),1);     
    current_solution = parameters;
    Eigen::VectorXd proposed_solution = current_solution;
    Eigen::VectorXd random_solution = current_solution;
    double current_misfit_value = common_functions::misfitFunctionEvaluation(run_model, current_solution, Np, Nr, Ng);
    double current_RMS_value = sqrt(current_misfit_value/no_data_points);
    double proposed_misfit_value = current_misfit_value;
    double proposed_RMS_value = current_RMS_value;
    double no_models = (max_Np + 1)*(max_Nr + 1)*(max_Ng + 1);
    int model_iterator = 0;
    int current_Np = Np;
    int current_Nr = Nr;
    int current_Ng = Ng;
    Np = 0;
    Nr = 0;
    Ng = 0;
    
    int j;
    for (Np = 0; Np < max_Np+1; ++Np)
    {
        for (Nr = 0; Nr < max_Nr+1; ++Nr)
        {
            for (Ng = 0; Ng < max_Ng+1; ++Ng)
            {
                model_iterator = Np*((max_Nr + 1)*(max_Ng + 1)) + Nr*(max_Ng + 1) + Ng;
                no_parameters = 1 + Np + 2*Nr + 3*Ng;
                proposed_solution.resize(no_parameters);
                random_solution.resize(no_parameters);
                while (current_resistivity_is_smaller)
                {
                    common_functions::makeRandomSolution(run_model,random_solution,min_parameter_values,max_parameter_values,initial_category_values,category_standard_deviation,initial_distribution,Np,Nr,Ng);
                    // fill proposed solution with initial values
                    i = 0;
                    proposed_solution(i) = 0.271;
                    for (j = 1; j < Np + 1; ++j)
                    {
                        ++i;
                        proposed_solution(i) = 1.0;
                    }
                    for (j = 1; j < Nr + 1; ++j)
                    {
                        ++i;
                        proposed_solution(i) = 1.0;
                    }
                    for (j = 1; j < Nr + 1; ++j)
                    {
                        ++i;
                        proposed_solution(i) = random_solution(i);
                    }
                    for (j = 1; j < Ng + 1; ++j)
                    {
                        ++i;
                        proposed_solution(i) = 1.0;
                    }
                    for (j = 1; j < Ng + 1; ++j)
                    {
                        ++i;
                        proposed_solution(i) = random_solution(i);
                    }
                    for (j = 1; j < Ng + 1; ++j)
                    {
                        ++i;
                        proposed_solution(i) = random_solution(i);
                    }
                    current_resistivity_min = common_functions::resistivityFunctionEvaluation(run_model, proposed_solution, run_model.get_temporary_dump_dir() + "/temp_profile_" + string_run_id, Np, Nr, Ng); 
                    current_resistivity_is_smaller = current_resistivity_min < min_resistivity_value;
                }
                current_resistivity_is_smaller = true;
                proposed_misfit_value = common_functions::misfitFunctionEvaluation(run_model, proposed_solution, Np, Nr, Ng);
                temp_bin.push_back(proposed_misfit_value);
                archive.push_back(temp_bin);
                temp_bin.clear();
                for (i = 0; i < no_parameters; ++i)
                {
                    temp_bin.push_back(proposed_solution(i));
                }
                best_solutions.push_back(temp_bin);
                temp_bin.clear();
                best_misfit_values.push_back(proposed_misfit_value);
            }
        }
    }

    all_misfit_values.push_back(current_misfit_value);
    latest_misfit_values.push_back(current_misfit_value);

    Eigen::MatrixXd all_latest_solution_values = Eigen::MatrixXd::Zero(no_iter_history,no_parameters);  // Just a dummy, not in use per now.
    Eigen::VectorXd all_latest_misfit_values = Eigen::MatrixXd::Zero(no_iter_history,1);

    common_functions::pushBackDoubleEigenVector(all_latest_misfit_values,current_misfit_value);
    
    // initial temperature
    if (temperature_schedule == 1)
    {
        chi_squared mydist(no_parameters);  // degrees of freedom = no_parameters
        double percentile = quantile(mydist,1 - this_run.get_p_chi());
        initial_temperature = 2*(current_misfit_value - min_misfit_value)/percentile;
    }
    double temperature = initial_temperature;
    
    // Create fitness vectors
    Eigen::VectorXd fitness_values = Eigen::MatrixXd::Zero(no_models,1);
    for (i = 0; i < no_models; ++i)
        fitness_values(i) = archive[i][0];

    Eigen::VectorXd accumulated_norm_fitness = Eigen::MatrixXd::Zero(fitness_values.size() + 1, 1);
    common_functions::cumulativeFitness(fitness_values, accumulated_norm_fitness);
    
    cout << "------------" << fitness_values << endl;
    cout << "------------" << accumulated_norm_fitness << "--------------" << endl;
    int proposed_Np = Np;
    int proposed_Nr = Nr;
    int proposed_Ng = Ng;

    bool update_temperature = false;
    double accept_prob = 0.0;
    double uniform_0_1;
    double latest_misfit_min;
    double latest_misfit_mean;
    double reduce_step_length_const;
    bool accept_proposed_when_smaller = false;
    int accept_proposed_when_smaller_counter = 0;
    int accept_proposed_counter = 0;
    double accept_ratio = 0;
    double p_model = 0;
    Eigen::MatrixXd::Index eigen_index;
    int current_model_index;
    int epoch_length = 0;

    // Local variables
    Eigen::VectorXd local_current_solution = current_solution;
    Eigen::VectorXd local_proposed_solution = proposed_solution;
    double local_current_misfit_value = current_misfit_value;
    double local_proposed_misfit_value = proposed_misfit_value;
    double local_proposed_RMS_value = proposed_RMS_value;
    double local_current_RMS_value = current_RMS_value;

    int total_iterations = 0;
    int global_iterator = 0;
    int local_iterator = 0;

    end_time = common_functions::getRealTime();
    std::string solution_filepath = run_model.get_siman_solution_dir() + "/" + run_model.get_solution_filename();
    std::ofstream out_stream(solution_filepath.c_str(), std::ios::out | std::ios::trunc);
    if (!out_stream.is_open())
        std::cout << "Unable to open file()" << endl;
    out_stream << "Algorithm: Transdimensional Simulated Annealing \n";
    out_stream << std::setw(7) << "Iter" << std::setw(12) << "Current rms" << std::setw(12) << "Local misfit" << std::setw(9) << "Local rms" << std::setw(9) << "Prop RMS";
    out_stream << std::setw(13) << "Accept prob" << std::setw(13) << "Accept ratio" << std::setw(13) << "Temperature" << std::setw(13) << "Time" << std::setw(3) << "n";
    out_stream << std::setw(4) << "Np" << std::setw(4) << "Nr" << std::setw(4) << "Ng\n";
    out_stream << std::setw(7) << total_iterations << std::setw(12) << current_RMS_value << current_misfit_value << std::setw(9) << current_RMS_value;
    out_stream << std::setw(9) << proposed_RMS_value << std::setw(13)   << accept_prob << std::setw(13) << accept_ratio << std::setw(13) << temperature << std::setw(13) << end_time - start_time << std::setw(3) << num_halves << std::setw(4) << Np << std::setw(4) << Nr << std::setw(4) << Ng;
    for (i = 0; i < current_solution.size(); ++i)
        out_stream << std::setw(15) << current_solution(i);
    out_stream << "\n";

    std::string variable_filepath = run_model.get_siman_variable_dir() + "/" + run_model.get_variable_filename();
    this_run.writeVariableValuesToTxtFile(run_model, parameters, min_parameter_values, max_parameter_values, variable_filepath, initial_filepath);
    
    srand(seed);
    rand();
    rand();
    double rand01_1 = (double) rand() / RAND_MAX; 
    int local_run_index = 0;
    bool no_change = true;
    bool increase_Np = false;
    bool decrease_Np = false;
    bool increase_Nr = false;
    bool decrease_Nr = false;
    bool increase_Ng = false;
    bool decrease_Ng = false;

    this_run.set_no_change(no_change);
    this_run.set_increase_Np(increase_Np);
    this_run.set_decrease_Np(decrease_Np);
    this_run.set_increase_Nr(increase_Nr);
    this_run.set_decrease_Nr(decrease_Nr);
    this_run.set_increase_Ng(increase_Ng);
    this_run.set_decrease_Ng(decrease_Ng);

    bool proceed = true;
    ++total_iterations;
    rand();
    rand();
    rand();
    while (proceed)
    {
        p_model = (double) rand() / RAND_MAX;
        if (p_model < best_model_prob)
        {
            rand01_1 = (double) rand() / RAND_MAX;
            for (i = 0; i < no_models; ++i)
            {
                if ( (accumulated_norm_fitness(i) <= rand01_1) && (rand01_1 < accumulated_norm_fitness(i+1)) )
                    current_model_index = i;
            }
        }
        else
        {
             current_model_index = common_functions::intUnifRand(0,no_models); 
        }
        common_functions::inverseModelIndex(current_model_index,Np,Nr,Ng,max_Np,max_Nr,max_Ng); 
        no_parameters = 1 + Np + 2*Nr + 3*Ng;
        local_current_solution.resize(no_parameters);
        for (i = 0; i < no_parameters; ++i)
            local_current_solution(i) = best_solutions[current_model_index][i];
        local_proposed_solution.resize(no_parameters);
        local_current_misfit_value = common_functions::misfitFunctionEvaluation(run_model, local_current_solution, Np, Nr, Ng);
        epoch_length = epoch_const*no_parameters;

        for (local_run_index = 0; local_run_index < epoch_length; ++local_run_index)
        {
            while (current_resistivity_is_smaller)
            { 
                rand01_1 = (double) rand() / RAND_MAX;
                this_run.newProposedSolution(run_model, local_proposed_solution, local_current_solution, min_parameter_values, max_parameter_values, all_latest_solution_values, all_latest_misfit_values, num_halves, global_iterator, proposed_Np, proposed_Nr, proposed_Ng, Np, Nr, Ng, min_step_length_values, max_step_length_values, initial_category_values, category_standard_deviation, initial_distribution);
                current_resistivity_min = common_functions::resistivityFunctionEvaluation(run_model, local_proposed_solution, run_model.get_temporary_dump_dir() + "/temp_profile_" + string_run_id, Np, Nr, Ng); 
                current_resistivity_is_smaller = current_resistivity_min < min_resistivity_value;
            }
            current_resistivity_is_smaller = true;

            local_proposed_misfit_value = common_functions::misfitFunctionEvaluation(run_model, local_proposed_solution, Np, Nr, Ng);
        
            common_functions::pushBackDoubleEigenVector(all_latest_misfit_values,local_proposed_misfit_value);
            if (local_proposed_misfit_value <= local_current_misfit_value)
            {
                local_current_solution = local_proposed_solution;
                local_current_misfit_value = local_proposed_misfit_value;
                accept_prob = 1.0;
                accept_proposed_counter++;
            }
            else
            {
                uniform_0_1 = ((double)rand()/(double)RAND_MAX);
                accept_prob = exp(- (local_proposed_misfit_value - local_current_misfit_value)/ temperature );
                if (uniform_0_1 < accept_prob)
                {
                    local_current_solution = local_proposed_solution;
                    local_current_misfit_value = local_proposed_misfit_value;
                    accept_proposed_counter++;
                }
            }
            if (local_current_misfit_value < best_misfit_values[current_model_index])
            {
                for (i = 0; i < no_parameters; ++i)
                    best_solutions[current_model_index][i] = local_current_solution(i);
            }

            archive[current_model_index].push_back(local_current_misfit_value);

            accept_ratio = (double)accept_proposed_counter/((double)total_iterations + 1.0);
            local_proposed_RMS_value = sqrt(local_proposed_misfit_value/no_data_points);
            local_current_RMS_value = sqrt(local_current_misfit_value/no_data_points);
            all_misfit_values.push_back(local_current_misfit_value);
            latest_misfit_values.push_back(local_current_misfit_value);
           
            cout << global_iterator << "_" << local_run_index << endl;
            cout << current_model_index << " : " << "(" << Np << "," << Nr << "," << Ng << ")" << endl;
            cout << "-----------------" << endl;
            cout << local_proposed_solution << endl;
            cout << "-----------------" << endl;
            cout << local_current_solution << endl;
            cout << "Run id:                  " << run_id << endl;
            cout << "Iteration number:        " << total_iterations << endl;
            cout << "Local current RMS value: " << local_current_RMS_value << endl << endl;
            
            ++total_iterations;
            
            end_time = common_functions::getRealTime();
            out_stream << std::setw(7) << total_iterations << std::setw(12) << current_RMS_value << std::setw(12) << local_current_misfit_value << std::setw(9) << local_current_RMS_value;
            out_stream << std::setw(9) << local_proposed_RMS_value << std::setw(13)   << accept_prob << std::setw(13) << accept_ratio << std::setw(13) << temperature << std::setw(13) << end_time - start_time << std::setw(3) << num_halves << std::setw(4) << Np << std::setw(4) << Nr << std::setw(4) << Ng;
            for (i = 0; i < local_current_solution.size(); ++i)
                out_stream << std::setw(15) << local_current_solution(i);
            out_stream << "\n";
        }
        update_temperature = true;

        if (update_temperature)
        {
            temperature = this_run.updateTemperature(run_model, all_misfit_values, min_misfit_value, no_parameters, temperature, initial_temperature, total_iterations);
            update_temperature = false;
        }

        if (this_run.get_reduce_step_length() && (local_iterator > this_run.get_num_iter_before_update()))
        {
            latest_misfit_min = *std::min_element(latest_misfit_values.begin(), latest_misfit_values.end());
            sum_latest_misfit_values = std::accumulate(latest_misfit_values.begin(), latest_misfit_values.end(), 0.0);
            latest_misfit_mean = sum_latest_misfit_values/latest_misfit_values.size();
            reduce_step_length_const = (latest_misfit_mean - latest_misfit_min)/latest_misfit_mean;
            if (reduce_step_length_const < this_run.get_update_threshold())
            {
                num_halves++;
                this_run.stepLengthReduction(min_step_length_values, max_step_length_values, 2.0, num_halves);
            }
            local_iterator = 0;
            latest_misfit_values.clear();
            latest_misfit_values.push_back(current_misfit_value);
            update_temperature = true;
        }

       
        // Terminate?

        if (this_run.get_max_iter_termination() && (total_iterations > this_run.get_max_iter() )) // this_run.get_max_iter() 
        {
            proceed = false;
            cout << "max iteration termination" << endl;
        }

        if (this_run.get_fn_value_termination())
        {
            if (!run_model.get_noise())
            {
                if ((current_RMS_value < 1.0))  
                {   
                    proceed = false;
                    cout << "silent function value termination" << endl;
                }
            }
            else
            {
                if ((current_RMS_value < 1.0))
                {
                    proceed = false;
                    cout << "noise function value termination" << endl;
                }
            }
        }

        if (num_halves > 5)
        {
            proceed = false;
            cout << "num_halves termination" << endl;
        }
        
        temp_archive.resize(archive[current_model_index].size());
        for (i = 0; i < temp_archive.size(); ++i)
            temp_archive(i) = archive[current_model_index][i];

        fitness_values(current_model_index) = common_functions::meanOfBestPartOfVector(temp_archive,20.0);
        common_functions::cumulativeFitness(fitness_values, accumulated_norm_fitness);
        
        global_iterator++;
        local_iterator++;

        
        if (local_current_misfit_value < current_misfit_value)
        {
            current_misfit_value = local_current_misfit_value;
            current_solution = local_current_solution;
            current_Np = Np;
            current_Nr = Nr;
            current_Ng = Ng;

        }
        else
        {
            uniform_0_1 = ((double)rand()/(double)RAND_MAX);
            accept_prob = exp(- (local_current_misfit_value - current_misfit_value)/ temperature );
            if (uniform_0_1 < accept_prob)
            {
                current_solution = local_current_solution;
                current_misfit_value = local_current_misfit_value;
                current_Np = Np;
                current_Nr = Nr;
                current_Ng = Ng;
            }
        }
        
        common_functions::resistivityFunctionEvaluation(run_model,current_solution,"synthetic_keep", current_Np, current_Nr, current_Ng);

        current_RMS_value = sqrt(current_misfit_value/no_data_points);

        cout << endl << "///////////////////////////////////" << endl;
        cout << local_current_solution << endl;
        cout << "-----------------" << endl;
        cout << current_solution << endl;
        cout << "Run id:                   " << run_id << endl;
        cout << "Iteration number:         " << total_iterations << endl;
        cout << "Global current RMS value: " << current_RMS_value << endl << endl;
        cout << endl << "///////////////////////////////////" << endl;



    }

    out_stream.close();


    end_time = common_functions::getRealTime();
    cout << "Elapsed time (s):  " << end_time - start_time << endl; 
    cout << "Solution in: " << solution_filepath << endl;
    cout << "Variable values in: " << variable_filepath << endl;

    return 0;
}
